import unittest
import numpy as np

from simsopt._core.dofs import get_owners, Dofs
from simsopt._core.optimizable import Target
from simsopt.objectives.functions import Identity, Adder, TestObject2, \
    Rosenbrock, Affine, Failer


class GetOwnersTests(unittest.TestCase):
    def test_no_dependents(self):
        """
        For an object that does not depend on anything, just return the
        original object.
        """
        obj = object()
        self.assertEqual(get_owners(obj), [obj])

        iden = Identity()
        self.assertEqual(get_owners(iden), [iden])

    def test_depth_1(self):
        """
        Check cases in which the original object depends on 1 or more others.
        """
        o1 = Identity()
        o2 = Identity()
        o1.o2 = o2
        o1.depends_on = ["o2"]
        self.assertEqual(get_owners(o1), [o1, o2])

        o3 = object()
        o1.depends_on = ["o3", "o2"]
        o1.o3 = o3
        self.assertEqual(get_owners(o1), [o1, o3, o2])

    def test_depth_2(self):
        """
        Check cases in which the original object depends on another, which
        depends on another.
        """
        o1 = Identity()
        o2 = Identity()
        o3 = object()
        o1.depends_on = ["o2"]
        o2.depends_on = ["o3"]
        o1.o2 = o2
        o2.o3 = o3
        self.assertEqual(get_owners(o1), [o1, o2, o3])

    def test_circular2(self):
        """
        Verify that a circular dependency among 2 objects is detected.
        """
        o1 = Identity()
        o2 = Identity()
        o1.depends_on = ["o2"]
        o2.depends_on = ["o1"]
        o1.o2 = o2
        o2.o1 = o1
        with self.assertRaises(RuntimeError):
            get_owners(o1)

    def test_circular3(self):
        """
        Verify that a circular dependency among 3 objects is detected.
        """
        o1 = Identity()
        o2 = Identity()
        o3 = Identity()
        o1.depends_on = ["o2"]
        o2.depends_on = ["o3"]
        o3.depends_on = ["o1"]
        o1.o2 = o2
        o2.o3 = o3
        o3.o1 = o1
        with self.assertRaises(RuntimeError):
            get_owners(o1)

    def test_circular4(self):
        """
        Verify that a circular dependency among 4 objects is detected.
        """
        o1 = Identity()
        o2 = Identity()
        o3 = Identity()
        o4 = Identity()
        o1.depends_on = ["o2"]
        o2.depends_on = ["o3"]
        o3.depends_on = ["o4"]
        o4.depends_on = ["o1"]
        o1.o2 = o2
        o2.o3 = o3
        o3.o4 = o4
        o4.o1 = o1
        with self.assertRaises(RuntimeError):
            get_owners(o1)


class DofsTests(unittest.TestCase):
    def test_no_dependents(self):
        """
        Tests for an object that does not depend on other objects.
        """
        obj = Adder(4)
        obj.set_dofs([101, 102, 103, 104])
        dofs = Dofs([obj.J])
        np.testing.assert_allclose(dofs.x, [101, 102, 103, 104])
        self.assertEqual(dofs.all_owners, [obj])
        self.assertEqual(dofs.dof_owners, [obj, obj, obj, obj])
        np.testing.assert_allclose(dofs.indices, [0, 1, 2, 3])
        dummy = dofs.f()  # f must be evaluated before we know nvals_per_func
        self.assertEqual(list(dofs.nvals_per_func), [1])
        self.assertEqual(dofs.nvals, 1)

        obj.fixed = [True, False, True, False]
        dofs = Dofs([obj.J])
        np.testing.assert_allclose(dofs.x, [102, 104])
        self.assertEqual(dofs.all_owners, [obj])
        self.assertEqual(dofs.dof_owners, [obj, obj])
        np.testing.assert_allclose(dofs.indices, [1, 3])

        obj.fixed[0] = False
        dofs = Dofs([obj.J])
        np.testing.assert_allclose(dofs.x, [101, 102, 104])
        self.assertEqual(dofs.all_owners, [obj])
        self.assertEqual(dofs.dof_owners, [obj, obj, obj])
        np.testing.assert_allclose(dofs.indices, [0, 1, 3])

    def test_no_fixed(self):
        """
        Test behavior when there is no 'fixed' attribute.
        """
        obj = Adder(4)
        del obj.fixed
        self.assertFalse(hasattr(obj, 'fixed'))
        obj.set_dofs([101, 102, 103, 104])
        dofs = Dofs([obj.J])
        np.testing.assert_allclose(dofs.x, [101, 102, 103, 104])
        self.assertEqual(dofs.all_owners, [obj])
        self.assertEqual(dofs.dof_owners, [obj, obj, obj, obj])
        np.testing.assert_allclose(dofs.indices, [0, 1, 2, 3])

    def test_with_dependents(self):
        """
        Test the case in which the original object depends on another object.
        """
        o1 = Adder(3)
        o2 = Adder(4)
        o1.set_dofs([10, 11, 12])
        o2.set_dofs([101, 102, 103, 104])
        o1.depends_on = ["o2"]
        o1.o2 = o2
        dofs = Dofs([o1.J])
        np.testing.assert_allclose(dofs.x, [10, 11, 12, 101, 102, 103, 104])
        self.assertEqual(dofs.all_owners, [o1, o2])
        self.assertEqual(dofs.dof_owners, [o1, o1, o1, o2, o2, o2, o2])
        np.testing.assert_allclose(dofs.indices, [0, 1, 2, 0, 1, 2, 3])
        f = dofs.f()  # f must be evaluated before we know nvals_per_func
        self.assertEqual(list(dofs.nvals_per_func), [1])
        self.assertEqual(dofs.nvals, 1)

        o1.fixed = [True, False, True]
        o2.fixed = [False, False, True, True]
        del o1.depends_on
        o2.depends_on = ["o1"]
        o2.o1 = o1
        dofs = Dofs([o2.J])
        np.testing.assert_allclose(dofs.x, [101, 102, 11])
        self.assertEqual(dofs.all_owners, [o2, o1])
        self.assertEqual(dofs.dof_owners, [o2, o2, o1])
        np.testing.assert_allclose(dofs.indices, [0, 1, 1])

    def test_vector_valued(self):
        """
        For a function that returns a vector rather than a scalar, make
        sure Dofs.f(), Dofs.jac(), and Dofs.fd_jac() behave correctly.
        """
        for nparams in range(1, 5):
            for nvals in range(1, 5):
                o = Affine(nparams=nparams, nvals=nvals)
                o.set_dofs((np.random.rand(nparams) - 0.5) * 4)
                dofs = Dofs([o], diff_method="centered")
                np.testing.assert_allclose(dofs.f(), np.matmul(o.A, o.x) + o.B, \
                                           rtol=1e-13, atol=1e-13)
                np.testing.assert_allclose(dofs.jac(), o.A, rtol=1e-13, atol=1e-13)
                np.testing.assert_allclose(dofs.fd_jac(), \
                                           o.A, rtol=1e-7, atol=1e-7)

    def test_multiple_vector_valued(self):
        """
        For a function that returns a vector rather than a scalar, make
        sure Dofs.f(), Dofs.jac(), and Dofs.fd_jac() behave correctly.
        """
        for nparams1 in range(1, 5):
            for nvals1 in range(1, 5):
                nparams2 = np.random.randint(1, 6)
                nparams3 = np.random.randint(1, 6)
                nvals2 = np.random.randint(1, 6)
                nvals3 = np.random.randint(1, 6)
                o1 = Affine(nparams=nparams1, nvals=nvals1)
                o2 = Affine(nparams=nparams2, nvals=nvals2)
                o3 = Affine(nparams=nparams3, nvals=nvals3)
                dofs = Dofs([o1, o2, o3], diff_method="centered")
                dofs.set((np.random.rand(nparams1 + nparams2 + nparams3) - 0.5) * 4)
                f1 = np.matmul(o1.A, o1.x) + o1.B
                f2 = np.matmul(o2.A, o2.x) + o2.B
                f3 = np.matmul(o3.A, o3.x) + o3.B
                np.testing.assert_allclose(dofs.f(), np.concatenate((f1, f2, f3)), \
                                           rtol=1e-13, atol=1e-13)
                true_jac = np.zeros((nvals1 + nvals2 + nvals3, nparams1 + nparams2 + nparams3))
                true_jac[0:nvals1, 0:nparams1] = o1.A
                true_jac[nvals1:nvals1 + nvals2, nparams1:nparams1 + nparams2] = o2.A
                true_jac[nvals1 + nvals2:nvals1 + nvals2 + nvals3, \
                         nparams1 + nparams2:nparams1 + nparams2 + nparams3] = o3.A
                np.testing.assert_allclose(dofs.jac(), true_jac, rtol=1e-13, atol=1e-13)
                np.testing.assert_allclose(dofs.fd_jac(), \
                                           true_jac, rtol=1e-7, atol=1e-7)

    def test_mixed_vector_valued(self):
        """
        For a mixture of functions that return a scalar vs return a
        vector, make sure Dofs.f(), Dofs.jac(), and Dofs.fd_jac()
        behave correctly.
        """
        for nparams1 in range(1, 5):
            for nvals1 in range(1, 5):
                nparams2 = np.random.randint(1, 6)
                nparams3 = np.random.randint(1, 6)
                nvals2 = np.random.randint(1, 6)
                nvals3 = np.random.randint(1, 6)
                o1 = Affine(nparams=nparams1, nvals=nvals1)
                o2 = Affine(nparams=nparams2, nvals=nvals2)
                o3 = Affine(nparams=nparams3, nvals=nvals3)
                a1 = Adder(n=2)
                a2 = Adder(n=3)
                dofs = Dofs([o1, o2, a1, o3, a2], diff_method="centered")
                dofs.set((np.random.rand(nparams1 + nparams2 + nparams3 + 5) - 0.5) * 4)
                f1 = np.matmul(o1.A, o1.x) + o1.B
                f2 = np.matmul(o2.A, o2.x) + o2.B
                f3 = np.array([a1.f])
                f4 = np.matmul(o3.A, o3.x) + o3.B
                f5 = np.array([a2.f])
                np.testing.assert_allclose(dofs.f(), np.concatenate((f1, f2, f3, f4, f5)), \
                                           rtol=1e-13, atol=1e-13)
                true_jac = np.zeros((nvals1 + nvals2 + nvals3 + 2, nparams1 + nparams2 + nparams3 + 5))
                true_jac[0:nvals1, 0:nparams1] = o1.A
                true_jac[nvals1:nvals1 + nvals2, nparams1:nparams1 + nparams2] = o2.A
                true_jac[nvals1 + nvals2:nvals1 + nvals2 + 1, \
                         nparams1 + nparams2:nparams1 + nparams2 + 2] = np.ones(2)
                true_jac[nvals1 + nvals2 + 1:nvals1 + nvals2 + 1 + nvals3, \
                         nparams1 + nparams2 + 2:nparams1 + nparams2 + 2 + nparams3] = o3.A
                true_jac[nvals1 + nvals2 + 1 + nvals3:nvals1 + nvals2 + nvals3 + 2, \
                         nparams1 + nparams2 + nparams3 + 2:nparams1 + nparams2 + nparams3 + 5] = np.ones(3)
                np.testing.assert_allclose(dofs.jac(), true_jac, rtol=1e-13, atol=1e-13)
                np.testing.assert_allclose(dofs.fd_jac(), \
                                           true_jac, rtol=1e-7, atol=1e-7)

    def test_Jacobian(self):
        for n in range(1, 20):
            v1 = np.random.rand() * 4 - 2
            v2 = np.random.rand() * 4 - 2
            o = TestObject2(v1, v2)
            o.adder.set_dofs(np.random.rand(2) * 4 - 2)
            o.t.set_dofs([np.random.rand() * 4 - 2])
            o.t.adder1.set_dofs(np.random.rand(3) * 4 - 2)
            o.t.adder2.set_dofs(np.random.rand(2) * 4 - 2)
            r = Rosenbrock(b=3.0)
            r.set_dofs(np.random.rand(2) * 3 - 1.5)
            a = Affine(nparams=3, nvals=3)

            # Randomly fix some of the degrees of freedom
            o.fixed = np.random.rand(2) > 0.5
            o.adder.fixed = np.random.rand(2) > 0.5
            o.t.adder1.fixed = np.random.rand(3) > 0.5
            o.t.adder2.fixed = np.random.rand(2) > 0.5
            r.fixed = np.random.rand(2) > 0.5
            a.fixed = np.random.rand(3) > 0.5

            rtol = 1e-6
            atol = 1e-6

            for j in range(4):
                # Try different sets of the objects:
                if j == 0:
                    dofs = Dofs([o.J, r.terms, o.t.J])
                    nvals = 4
                    nvals_per_func = [1, 2, 1]
                elif j == 1:
                    dofs = Dofs([r.term2, r.terms])
                    nvals = 3
                    nvals_per_func = [1, 2]
                elif j == 2:
                    dofs = Dofs([r.term2, Target(o.t, 'f'), r.term1, Target(o, 'f')])
                    nvals = 4
                    nvals_per_func = [1, 1, 1, 1]
                elif j == 3:
                    dofs = Dofs([a, o])
                    nvals = 4
                    nvals_per_func = [3, 1]

                jac = dofs.jac()
                dofs.diff_method = "forward"
                fd_jac = dofs.fd_jac()
                dofs.diff_method = "centered"
                fd_jac_centered = dofs.fd_jac()
                #print('j=', j, '  Diff in Jacobians:', jac - fd_jac)
                #print('jac: ', jac)
                #print('fd_jac: ', fd_jac)
                #print('fd_jac_centered: ', fd_jac_centered)
                #print('shapes: jac=', jac.shape, '  fd_jac=', fd_jac.shape, '  fd_jac_centered=', fd_jac_centered.shape)
                np.testing.assert_allclose(jac, fd_jac, rtol=rtol, atol=atol)
                np.testing.assert_allclose(fd_jac, fd_jac_centered, rtol=rtol, atol=atol)
                self.assertEqual(dofs.nvals, nvals)
                self.assertEqual(list(dofs.nvals_per_func), nvals_per_func)

    def test_failures(self):
        """
        Verify that if ObjectiveFailure is raised during function
        evaluations, a vector is returned filled with the expected
        number.
        """
        nvals = 3
        fail_val = 1.0e8
        o1 = Failer(nvals=nvals)
        d1 = Dofs([o1], fail=fail_val)
        # First eval should not fail:
        f = d1.f()
        np.testing.assert_allclose(f, np.full(nvals, 1.0))
        # There should be a failure on the 2nd evaluation:
        f = d1.f()
        np.testing.assert_allclose(f, np.full(nvals, fail_val))
        # Third eval should not fail:
        f = d1.f()
        np.testing.assert_allclose(f, np.full(nvals, 1.0))

        # Try an example with >1 object in the dofs, and with NaN
        # instead of a finite value for the failure value.
        fail_val = np.NAN
        o2 = Failer(nvals=3)
        r2 = Rosenbrock()
        d2 = Dofs([o2, r2.terms], fail=fail_val)
        # First eval should not fail:
        f = d2.f()
        np.testing.assert_allclose(f, [1., 1., 1., -1., 0.])
        # There should be a failure on the 2nd evaluation:
        f = d2.f()
        np.testing.assert_allclose(f, np.full(5, fail_val))
        # Third eval should not fail:
        f = d2.f()
        np.testing.assert_allclose(f, [1., 1., 1., -1., 0.])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import logging
import numpy as np

from simsopt.mhd import Spec
from simsopt import LeastSquaresProblem
from simsopt import least_squares_serial_solve
import os

"""
This script implements the "2DOF_vmecOnly_targetIotaAndVolume" example from
https://github.com/landreman/stellopt_scenarios

This optimization problem has two independent variables, representing
the helical shape of the magnetic axis. The problem also has two
objectives: the plasma volume and the rotational transform on the
magnetic axis.

The resolution in this example (i.e. ns, mpol, and ntor) is somewhat
lower than in the stellopt_scenarios version of the example, just so
this example runs fast.

Details of the optimum and a plot of the objective function landscape
can be found here:
https://github.com/landreman/stellopt_scenarios/tree/master/2DOF_vmecOnly_targetIotaAndVolume
"""

# This next line turns on detailed logging. It can be commented out if
# you do not want such verbose output.
logging.basicConfig(level=logging.INFO)

# Initialize SPEC from an input file
equil = Spec(os.path.join(os.path.dirname(__file__), 'inputs', '2DOF_targetIotaAndVolume.sp'))

# If the xspec executable is not in PATH, the path to the executable
# should be specified as in the following line:
# equil = Spec('2DOF_targetIotaAndVolume.sp', exe='/Users/mattland/SPEC/xspec')

surf = equil.boundary

# VMEC parameters are all fixed by default, while surface parameters are all non-fixed by default.
# You can choose which parameters are optimized by setting their 'fixed' attributes.
surf.all_fixed()
surf.set_fixed('rc(1,1)', False)
surf.set_fixed('zs(1,1)', False)

# Each Target is then equipped with a shift and weight, to become a
# term in a least-squares objective function.  A list of terms are
# combined to form a nonlinear-least-squares problem.
desired_volume = 0.15
volume_weight = 1
term1 = (equil.volume, desired_volume, volume_weight)

desired_iota = -0.41
iota_weight = 1
term2 = (equil.iota, desired_iota, iota_weight)

prob = LeastSquaresProblem([term1, term2])

# Solve the minimization problem:
least_squares_serial_solve(prob, grad=True)

print("At the optimum,")
print(" rc(m=1,n=1) = ", surf.get_rc(1, 1))
print(" zs(m=1,n=1) = ", surf.get_zs(1, 1))
print(" volume, according to VMEC    = ", equil.volume())
print(" volume, according to Surface = ", surf.volume())
print(" iota on axis = ", equil.iota())
print(" objective function = ", prob.objective())

# The tests here are based on values from the VMEC version in
# https://github.com/landreman/stellopt_scenarios/tree/master/2DOF_vmecOnly_targetIotaAndVolume
# Due to this and the fact that we don't yet have iota on axis from SPEC, the tolerances are wide.
assert np.abs(surf.get_rc(1, 1) - 0.0313066948) < 0.001
assert np.abs(surf.get_zs(1, 1) - (-0.031232391)) < 0.001
assert np.abs(equil.volume() - 0.178091) < 0.001
assert np.abs(surf.volume() - 0.178091) < 0.001
assert np.abs(equil.iota() - (-0.4114567)) < 0.001
assert (prob.objective() - 7.912501330E-04) < 0.2e-4



import unittest
import numpy as np
from simsopt._core.graph_optimizable import Optimizable
from simsopt._core.derivative import Derivative


class Opt(Optimizable):
    """
    This class defines a minimal object that can be optimized.
    """

    def __init__(self, n=3, x0=None):
        self.n = n
        x = x0 if x0 is not None else np.ones((n, ))
        super().__init__(x)

    def foo(self):
        return np.sin(self.local_full_x) * np.sum(self.local_full_x)

    def dfoo_vjp(self, v):
        n = self.n
        full_jac = np.zeros((n, n))
        for i in range(n):
            full_jac[i, i] = np.cos(self.local_full_x[i]) * np.sum(self.local_full_x)
            full_jac[:, i] += np.sin(self.local_full_x)
        return Derivative({self: v.T @ full_jac})


class InterSum(Optimizable):

    def __init__(self, opt_a, opt_b):
        Optimizable.__init__(self, x0=np.asarray([]), depends_on=[opt_a, opt_b])
        self.opt_a = opt_a
        self.opt_b = opt_b

    def bar(self):
        return np.sum(self.opt_a.foo()**2) + np.sum(self.opt_b.foo()**2)

    def dbar_vjp(self, v):
        return v[0] * self.opt_a.dfoo_vjp(2*self.opt_a.foo()) \
            + v[0] * self.opt_b.dfoo_vjp(2*self.opt_b.foo())


class InterProd(Optimizable):

    def __init__(self, opt_a, opt_b):
        Optimizable.__init__(self, x0=np.asarray([]), depends_on=[opt_a, opt_b])
        self.opt_a = opt_a
        self.opt_b = opt_b

    def bar(self):
        return np.prod(self.opt_a.foo()) * np.prod(self.opt_b.foo())

    def dbar_vjp(self, v):
        n_a = self.opt_a.n
        v_a = np.zeros((n_a, ))
        for i in range(n_a):
            v_a[i] = np.prod(self.opt_a.foo())/self.opt_a.foo()[i]
        v_a *= np.prod(self.opt_b.foo())

        n_b = self.opt_b.n
        v_b = np.zeros((n_b, ))
        for i in range(n_b):
            v_b[i] = np.prod(self.opt_b.foo())/self.opt_b.foo()[i]
        v_b *= np.prod(self.opt_a.foo())

        return v[0] * self.opt_a.dfoo_vjp(v_a) \
            + v[0] * self.opt_b.dfoo_vjp(v_b) \



class Obj(Optimizable):

    def __init__(self, inter_a, inter_b):
        Optimizable.__init__(self, x0=np.asarray([]), depends_on=[inter_a, inter_b])
        self.inter_a = inter_a
        self.inter_b = inter_b

    def J(self):
        return self.inter_a.bar() * self.inter_b.bar()

    def dJ(self):
        return self.inter_a.dbar_vjp([self.inter_b.bar()]) + self.inter_b.dbar_vjp([self.inter_a.bar()])


class DerivativeTests(unittest.TestCase):

    def test_taylor_graph(self):
        np.random.seed(1)
        # built a reasonably complex graph of two inputs, that both feed into
        # two intermediary results and then are combined into a final result.
        # i.e. f(g(x, y), h(x, y))

        opt1 = Opt(n=3)
        opt2 = Opt(n=2)

        intersum = InterSum(opt1, opt2)
        interprod = InterProd(opt1, opt2)
        obj = Obj(intersum, interprod)
        x = obj.x + 0.05 * np.random.standard_normal(size=obj.x.shape)
        obj.x = x
        x = obj.x
        h = np.random.standard_normal(size=x.shape)
        f = obj.J()
        df = obj.dJ()
        dfh = np.sum(df(obj)*h)
        err_old = 1e9
        for i in range(5, 11):
            eps = 0.5**i
            obj.x = x + 3 * eps * h
            fppp = obj.J()
            obj.x = x + 2 * eps * h
            fpp = obj.J()
            obj.x = x + eps * h
            fp = obj.J()
            obj.x = x - eps * h
            fm = obj.J()
            obj.x = x - 2 * eps * h
            fmm = obj.J()
            obj.x = x - 3 * eps * h
            fmmm = obj.J()
            # print(np.abs((fp-fm)/(2*eps) - dfh))
            dfhest = ((1/12) * fmm - (2/3) * fm + (2/3) * fp - (1/12) * fpp)/eps
            err = np.abs(dfhest - dfh)
            assert err < (0.6)**4 * err_old
            print(err_old/err)
            err_old = err

            # dfhest = ((-1/60)*fmmm + (3/20)*fmm -(3/4)*fm+(3/4)*fp-(3/20)*fpp + (1/60)*fppp)/eps
            # err = np.abs(dfhest - dfh)
            # print(err_old/err)
            # err_old = err

    def test_add_mul(self):
        opt1 = Opt(n=3)
        opt2 = Opt(n=2)

        dj1 = opt1.dfoo_vjp(np.ones(3))
        dj2 = opt2.dfoo_vjp(np.ones(2))

        dj1p2 = dj1 + dj2 + 2*dj1
        assert np.allclose(dj1p2(opt1), 3*dj1(opt1))
        assert np.allclose(dj1p2(opt2), dj2(opt2))

    def test_iadd_imul(self):
        opt1 = Opt(n=3)
        opt2 = Opt(n=2)

        dj1 = opt1.dfoo_vjp(np.ones(3))
        dj1_ = opt1.dfoo_vjp(np.ones(3))
        dj2 = opt2.dfoo_vjp(np.ones(2))

        dj1 += dj2
        assert np.allclose(dj1(opt2), dj2(opt2))
        dj1 += dj1_
        assert np.allclose(dj1(opt1), 2*dj1_(opt1))
        dj1 *= 1.5
        assert np.allclose(dj1(opt2), 1.5*dj2(opt2))
        assert np.allclose(dj1(opt1), 3*dj1_(opt1))

    def test_zero_when_not_found(self):
        opt1 = Opt(n=3)
        opt2 = Opt(n=2)

        dj1 = opt1.dfoo_vjp(np.ones(3))
        assert np.allclose(dj1(opt2), np.zeros((2, )))



#!/usr/bin/env python

import os
import numpy as np
from simsopt.mhd.vmec import Vmec
from simsopt.mhd.vmec_diagnostics import IotaTargetMetric
from scipy.optimize import minimize
from simsopt._core.util import ObjectiveFailure
from simsopt._core.dofs import Dofs
import sys

"""
Here, we perform an optimization begining with a 3 field period rotating ellipse
boundary to obtain iota = 0.381966. The derivatives are obtained with an adjoint
method. This is based on a published result in
Paul, Landreman, and Antonsen, Journal of Plasma Physics (2021). The number of
modes in the optimization space is slowly increased from |m|,|n| <= 2 to 5. At
the end, the initial and final profiles are plotted.
"""

# check whether we're in CI, in that case we make the run a bit cheaper
ci = "CI" in os.environ and os.environ['CI'].lower() in ['1', 'true']

# target profile of rotational transform, here just a constant:
target_function = lambda s: 0.381966
adjoint_epsilon = 1.e-1  # perturbation amplitude for adjoint solve

filename = os.path.join(os.path.dirname(__file__), 'inputs', 'input.rotating_ellipse')
vmec = Vmec(filename, ntheta=100, nphi=100)

# Lower resolution and function evaluations if in CI
if ci:
    vmec.indata.mpol = 6
    vmec.indata.ntor = 6
    vmec.indata.ns_array = np.zeros_like(vmec.indata.ns_array)
    vmec.indata.ns_array[0] = 9
    vmec.indata.ns_array[0] = 49
    maxres = 4
    maxfun = 1
else:
    maxres = 6
    maxfun = 15000
    import matplotlib.pyplot as plt

vmec.run()
iotas_init = vmec.wout.iotas

obj = IotaTargetMetric(vmec, target_function, adjoint_epsilon)

surf = vmec.boundary
surf.all_fixed(True)
# Slowly increase range of modes in optimization space
for max_mode in range(3, maxres):
    surf.fixed_range(mmin=0, mmax=max_mode,
                     nmin=-max_mode, nmax=max_mode, fixed=False)

    myfunc = Dofs([obj])

    # Define objective function and derivative that handle ObjectiveFailure
    def J(dofs):
        dofs_prev = myfunc.x
        try:
            myfunc.set(dofs)
            return myfunc.f()[0], np.squeeze(myfunc.jac())
        except ObjectiveFailure:
            myfunc.set(dofs_prev)
            return 2*myfunc.f()[0], 2*np.squeeze(myfunc.jac())

    res = minimize(
        fun=J, x0=myfunc.x, jac=True, method='L-BFGS-B',
        options={'maxfun': maxfun, 'ftol': 1e-8, 'gtol': 1e-8})
    print(f"max_mode={max_mode:d}  res={res['fun']:.3f}, jac={np.linalg.norm(res['jac']):.3f}")

    # Preserve the output file from the last iteration, so it is not
    # deleted when vmec runs again:
    vmec.files_to_delete = []

vmec.run()
iotas_final = vmec.wout.iotas

if not ci:
    # Plot result
    plt.figure()
    plt.plot(vmec.s_half_grid, iotas_init[1:], color='green')
    plt.plot(vmec.s_half_grid, iotas_final[1:], color='red')
    plt.axhline(target_function(0), color='blue')
    plt.legend(['Initial', 'Final', 'Target'])
    plt.xlabel(r'$s$')
    plt.ylabel(r'$\iota$')
    plt.show()



import unittest
import logging

import numpy as np
try:
    from mpi4py import MPI
except ImportError:
    MPI = None

from simsopt.objectives.functions import Identity, Rosenbrock, Beale
from simsopt._core.optimizable import Target
from simsopt.objectives.least_squares import LeastSquaresProblem, \
    LeastSquaresTerm
from simsopt.solve.serial import least_squares_serial_solve, serial_solve
if MPI is not None:
    from simsopt.util.mpi import MpiPartition
    from simsopt.solve.mpi import least_squares_mpi_solve


def mpi_solve_1group(prob, **kwargs):
    least_squares_mpi_solve(prob, MpiPartition(ngroups=1), **kwargs)


solvers = [serial_solve, least_squares_serial_solve]
if MPI is not None:
    solvers.append(mpi_solve_1group)

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class LeastSquaresProblemTests(unittest.TestCase):

    def test_solve_quadratic(self):
        """
        Minimize f(x,y,z) = 1 * (x - 1) ^ 2 + 2 * (y - 2) ^ 2 + 3 * (z - 3) ^ 2.
        The optimum is at (x,y,z)=(1,2,3), and f=0 at this point.
        """
        for solver in solvers:
            iden1 = Identity()
            iden2 = Identity()
            iden3 = Identity()
            term1 = (iden1.J, 1, 1)
            term2 = (iden2.J, 2, 2)
            term3 = (iden3.J, 3, 3)
            prob = LeastSquaresProblem([term1, term2, term3])
            if solver == serial_solve:
                solver(prob, tol=1e-12)
            else:
                solver(prob)
            self.assertAlmostEqual(prob.objective(), 0)
            self.assertAlmostEqual(iden1.x, 1)
            self.assertAlmostEqual(iden2.x, 2)
            self.assertAlmostEqual(iden3.x, 3)

    def test_solve_quadratic_fixed(self):
        """
        Same as test_solve_quadratic, except with different weights and x
        and z are fixed, so only y is optimized.
        """
        for solver in solvers:
            iden1 = Identity()
            iden2 = Identity()
            iden3 = Identity()
            iden1.x = 4
            iden2.x = 5
            iden3.x = 6
            iden1.names = ['x1']
            iden2.names = ['x2']
            iden3.names = ['x3']
            iden1.fixed = [True]
            iden3.fixed = [True]
            term1 = (iden1.J, 1, 1)
            term2 = (iden2.J, 2, 1 / 4.)
            term3 = (iden3.J, 3, 1 / 9.)
            prob = LeastSquaresProblem([term1, term2, term3])
            solver(prob)
            self.assertAlmostEqual(prob.objective(), 10)
            self.assertAlmostEqual(iden1.x, 4)
            self.assertAlmostEqual(iden2.x, 2)
            self.assertAlmostEqual(iden3.x, 6)

    def test_solve_quadratic_fixed_supplying_objects(self):
        """
        Same as test_solve_quadratic_fixed, except supplying objects
        rather than functions as targets.
        """
        for solver in solvers:
            iden1 = Identity()
            iden2 = Identity()
            iden3 = Identity()
            iden1.x = 4
            iden2.x = 5
            iden3.x = 6
            iden1.names = ['x1']
            iden2.names = ['x2']
            iden3.names = ['x3']
            iden1.fixed = [True]
            iden3.fixed = [True]
            term1 = [iden1, 1, 1]
            term2 = [iden2, 2, 1 / 4.]
            term3 = [iden3, 3, 1 / 9.]
            prob = LeastSquaresProblem([term1, term2, term3])
            solver(prob)
            self.assertAlmostEqual(prob.objective(), 10)
            self.assertAlmostEqual(iden1.x, 4)
            self.assertAlmostEqual(iden2.x, 2)
            self.assertAlmostEqual(iden3.x, 6)

    def test_solve_quadratic_fixed_supplying_attributes(self):
        """
        Same as test_solve_quadratic_fixed, except supplying attributes
        rather than functions as targets.
        """
        for solver in solvers:
            iden1 = Identity()
            iden2 = Identity()
            iden3 = Identity()
            iden1.x = 4
            iden2.x = 5
            iden3.x = 6
            iden1.names = ['x1']
            iden2.names = ['x2']
            iden3.names = ['x3']
            iden1.fixed = [True]
            iden3.fixed = [True]
            # Try a mix of explicit LeastSquaresTerms and tuples
            term1 = LeastSquaresTerm(Target(iden1, 'x'), 1, 1)
            term2 = (iden2, 'x', 2, 1 / 4.)
            term3 = (iden3, 'x', 3, 1 / 9.)
            prob = LeastSquaresProblem([term1, term2, term3])
            solver(prob)
            self.assertAlmostEqual(prob.objective(), 10)
            self.assertAlmostEqual(iden1.x, 4)
            self.assertAlmostEqual(iden2.x, 2)
            self.assertAlmostEqual(iden3.x, 6)

    def test_solve_quadratic_fixed_supplying_properties(self):
        """
        Same as test_solve_quadratic_fixed, except supplying @properties
        rather than functions as targets.
        """
        for solver in solvers:
            iden1 = Identity()
            iden2 = Identity()
            iden3 = Identity()
            iden1.x = 4
            iden2.x = 5
            iden3.x = 6
            iden1.names = ['x1']
            iden2.names = ['x2']
            iden3.names = ['x3']
            iden1.fixed = [True]
            iden3.fixed = [True]
            # Try a mix of explicit LeastSquaresTerms and lists
            term1 = [iden1, 'f', 1, 1]
            term2 = [iden2, 'f', 2, 1 / 4.]
            term3 = LeastSquaresTerm.from_sigma(Target(iden3, 'f'), 3, sigma=3)
            prob = LeastSquaresProblem([term1, term2, term3])
            solver(prob)
            self.assertAlmostEqual(prob.objective(), 10)
            self.assertAlmostEqual(iden1.x, 4)
            self.assertAlmostEqual(iden2.x, 2)
            self.assertAlmostEqual(iden3.x, 6)

    def test_solve_rosenbrock_using_scalars(self):
        """
        Minimize the Rosenbrock function using two separate least-squares
        terms.
        """
        for solver in solvers:
            for grad in [True, False]:
                r = Rosenbrock()
                term1 = (r.term1, 0, 1)
                term2 = (r.term2, 0, 1)
                prob = LeastSquaresProblem((term1, term2))
                if solver == serial_solve:
                    if grad == True:
                        continue
                    else:
                        solver(prob, tol=1e-12)
                else:
                    solver(prob, grad=grad)
                self.assertAlmostEqual(prob.objective(), 0)
                v = r.get_dofs()
                self.assertAlmostEqual(v[0], 1)
                self.assertAlmostEqual(v[1], 1)

    def test_solve_rosenbrock_using_vector(self):
        """
        Minimize the Rosenbrock function using a single vector-valued
        least-squares term.
        """
        for solver in solvers:
            for grad in [True, False]:
                r = Rosenbrock()
                prob = LeastSquaresProblem([(r.terms, 0, 1)])
                if solver == serial_solve:
                    if grad == True:
                        continue
                    else:
                        solver(prob, tol=1e-12)
                else:
                    solver(prob, grad=grad)
                self.assertAlmostEqual(prob.objective(), 0)
                v = r.get_dofs()
                self.assertAlmostEqual(v[0], 1)
                self.assertAlmostEqual(v[1], 1)

    def test_solve_with_finite_differences(self):
        """
        Minimize a function for which analytic derivatives are not
        provided. Provides test coverage for the finite-differencing
        options.
        """
        #for solver in [least_squares_serial_solve]:
        for solver in solvers:
            if solver == serial_solve:
                continue
            for abs_step in [0, 1.0e-7]:
                rel_steps = [0, 1.0e-7]
                if abs_step == 0:
                    rel_steps = [1.0e-7]
                for rel_step in rel_steps:
                    for diff_method in ["forward", "centered"]:
                        logger.debug(f'solver={solver} diff_method={diff_method} ' \
                                     f'abs_step={abs_step} rel_step={rel_step}')
                        b = Beale()
                        b.set_dofs([0.1, -0.2])
                        prob = LeastSquaresProblem([(b, 0, 1)], diff_method=diff_method,
                                                   abs_step=abs_step, rel_step=rel_step)
                        #least_squares_serial_solve(prob, grad=True)
                        solver(prob, grad=True)
                        np.testing.assert_allclose(prob.x, [3, 0.5])
                        np.testing.assert_allclose(prob.f(), [0, 0, 0], atol=1e-10)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

from simsopt.geo.surfacerzfourier import SurfaceRZFourier
from simsopt import LeastSquaresProblem
from simsopt import least_squares_serial_solve
"""
Optimize the minor radius and elongation of an axisymmetric torus to
obtain a desired volume and area.
"""

desired_volume = 0.6
desired_area = 8.0

# Start with a default surface, which is axisymmetric with major
# radius 1 and minor radius 0.1.
surf = SurfaceRZFourier()

# Parameters are all non-fixed by default, meaning they will be
# optimized.  You can choose to exclude any subset of the variables
# from the space of independent variables by setting their 'fixed'
# property to True.
surf.set_fixed('rc(0,0)')

# Each target function is then equipped with a shift and weight, to
# become a term in a least-squares objective function
term1 = (surf.volume, desired_volume, 1)
term2 = (surf.area, desired_area, 1)

# A list of terms are combined to form a nonlinear-least-squares
# problem.
prob = LeastSquaresProblem([term1, term2])

# Solve the minimization problem:
least_squares_serial_solve(prob)

print("At the optimum,")
print(" rc(m=1,n=0) = ", surf.get_rc(1, 0))
print(" zs(m=1,n=0) = ", surf.get_zs(1, 0))
print(" volume = ", surf.volume())
print(" area = ", surf.area())
print(" objective function = ", prob.objective())



#!/usr/bin/env python3

import numpy as np
from simsopt.geo.curverzfourier import CurveRZFourier
from simsopt.geo.curveobjectives import CurveLength
from simsopt.objectives.graph_least_squares import LeastSquaresProblem
from simsopt.solve.graph_serial import least_squares_serial_solve

"""
Minimize the length of a curve, holding the 0-frequency Fourier mode fixed.
The result should be a circle.
"""

# Create a curve:
nquadrature = 100
nfourier = 4
nfp = 5  # exploit rotational symmetry and only consider a fifth of the curve
curve = CurveRZFourier(nquadrature, nfourier, nfp, True)

# Initialize the Fourier amplitudes to some random values
x0 = np.random.rand(curve.dof_size) - 0.5
x0[0] = 3.0
curve.x = x0
print('Initial curve dofs: ', curve.x)

# Tell the curve object that the first Fourier mode is fixed, whereas
# all the other dofs are not.
curve.fix(0)

# The length objective is a separate object rather than a function of
# Curve itself.
obj = CurveLength(curve)

print('Initial curve length: ', obj.J())

# Each target function is then equipped with a shift and weight, to
# become a term in a least-squares objective function.
# A list of terms are combined to form a nonlinear-least-squares
# problem.
prob = LeastSquaresProblem.from_tuples([(obj.J, 0.0, 1.0)])

# At the initial condition, get the Jacobian two ways: analytic
# derivatives and finite differencing. The difference should be small.
# fd_jac = prob.dofs.fd_jac()
# jac = prob.dofs.jac()
# print('finite difference Jacobian:')
# print(fd_jac)
# print('Analytic Jacobian:')
# print(jac)
# print('Difference:')
# print(fd_jac - jac)

# Solve the minimization problem:
least_squares_serial_solve(prob)

print('At the optimum, x: ', prob.x)
print(' Final curve dofs: ', curve.local_full_x)
print(' Final curve length:    ', obj.J())
print(' Expected final length: ', 2 * np.pi * x0[0])
print(' objective function: ', prob.objective())



import numpy as np
from simsopt.geo.surfacexyzfourier import SurfaceXYZFourier
from simsopt.geo.surfacerzfourier import SurfaceRZFourier
from simsopt.geo.surfacexyztensorfourier import SurfaceXYZTensorFourier
from simsopt.geo.curverzfourier import CurveRZFourier
from simsopt.geo.curvexyzfourier import CurveXYZFourier
from simsopt.field.coil import Current
from pathlib import Path
TEST_DIR = (Path(__file__).parent / ".." / "test_files").resolve()


def get_surface(surfacetype, stellsym, phis=None, thetas=None, mpol=5, ntor=5,
                nphi=None, ntheta=None, full=False):
    nfp = 3
    if nphi is None:
        nphi = 11 if surfacetype == "SurfaceXYZTensorFourier" else 15
    if ntheta is None:
        ntheta = 11 if surfacetype == "SurfaceXYZTensorFourier" else 15

    if phis is None:
        phis = np.linspace(0, 1/nfp, nphi, endpoint=False)
    if thetas is None:
        if (surfacetype == "SurfaceXYZTensorFourier" or full == True):
            thetas = np.linspace(0, 1, ntheta, endpoint=False)
        else:
            thetas = np.linspace(0, 1/(1. + int(stellsym)), ntheta, endpoint=False)

    if surfacetype == "SurfaceXYZFourier":
        s = SurfaceXYZFourier(mpol=mpol, ntor=ntor, nfp=nfp, stellsym=stellsym,
                              quadpoints_phi=phis, quadpoints_theta=thetas)
    elif surfacetype == "SurfaceRZFourier":
        s = SurfaceRZFourier(mpol=mpol, ntor=ntor, nfp=nfp, stellsym=stellsym,
                             quadpoints_phi=phis, quadpoints_theta=thetas)
    elif surfacetype == "SurfaceXYZTensorFourier":
        s = SurfaceXYZTensorFourier(mpol=mpol, ntor=ntor, nfp=nfp, stellsym=stellsym,
                                    clamped_dims=[False, False, False],
                                    quadpoints_phi=phis, quadpoints_theta=thetas
                                    )
    else:
        raise Exception("surface type not implemented")
    return s


def get_exact_surface():
    filename_X = TEST_DIR / 'NCSX_test_data'/'X.dat'
    filename_Y = TEST_DIR / 'NCSX_test_data'/'Y.dat'
    filename_Z = TEST_DIR / 'NCSX_test_data'/'Z.dat'
    X = np.loadtxt(filename_X)
    Y = np.loadtxt(filename_Y)
    Z = np.loadtxt(filename_Z)
    xyz = np.concatenate((X[:, :, None], Y[:, :, None], Z[:, :, None]), axis=2)
    ntor = 16
    mpol = 10

    nfp = 1
    stellsym = False
    nphi = 33
    ntheta = 21

    phis = np.linspace(0, 1, nphi, endpoint=False)
    thetas = np.linspace(0, 1, ntheta, endpoint=False)
    s = SurfaceXYZFourier(mpol=mpol, ntor=ntor, nfp=nfp, stellsym=stellsym,
                          quadpoints_phi=phis, quadpoints_theta=thetas)
    s.least_squares_fit(xyz)

    return s



import logging
import unittest

import numpy as np
try:
    from mpi4py import MPI
except:
    MPI = None

from simsopt._core.graph_optimizable import Optimizable
from simsopt.objectives.graph_least_squares import LeastSquaresProblem
if MPI is not None:
    from simsopt.util.mpi import MpiPartition
    from simsopt.solve.graph_mpi import fd_jac_mpi, least_squares_mpi_solve

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class TestFunction1(Optimizable):
    def __init__(self):
        x = np.array([1.2, 0.9, -0.4])
        fixed = np.full(3, False)
        super().__init__(x0=x, fixed=fixed)

    def J(self):
        return np.exp(self.full_x[0] ** 2 - np.exp(self.full_x[1]) \
                      + np.sin(self.full_x[2]))

    return_fn_map = {'J': J}


class TestFunction2(Optimizable):
    def __init__(self):
        x = np.array([1.2, 0.9])
        fixed = np.full(2, False)
        super().__init__(x0=x, fixed=fixed)

    def f0(self):
        return np.exp(0 + self.full_x[0] ** 2 - np.exp(self.full_x[1]))

    def f1(self):
        return np.exp(1 + self.full_x[0] ** 2 - np.exp(self.full_x[1]))

    def f2(self):
        return np.exp(2 + self.full_x[0] ** 2 - np.exp(self.full_x[1]))

    def f3(self):
        return np.exp(3 + self.full_x[0] ** 2 - np.exp(self.full_x[1]))

    return_fn_map = {'f0': f0, 'f1': f1, 'f2': f2, 'f3': f3}


class TestFunction3(Optimizable):
    """
    This is the Rosenbrock function again, but with some unnecessary
    MPI communication added in order to test optimization with MPI.
    """

    def __init__(self, comm):
        x = [0., 0.]
        self.comm = comm
        self.dummy = 42
        logger.debug("inside test function 3 init")
        super().__init__(x0=x)

    def f0(self):
        # Do some random MPI stuff just for the sake of testing.
        self.comm.barrier()
        self.comm.bcast(self.full_x)
        return self.full_x[0] - 1

    def f1(self):
        # Do some random MPI stuff just for the sake of testing.
        self.comm.bcast(self.dummy)
        self.comm.barrier()
        return self.full_x[0] ** 2 - self.full_x[1]

    return_fn_map = {'f0': f0, 'f1': f1}


@unittest.skipIf(MPI is None, "Requires mpi4py")
class MpiPartitionTests(unittest.TestCase):
    def test_ngroups1(self):
        """
        Verify that all quantities make sense when ngroups = 1.
        """
        rank_world = MPI.COMM_WORLD.Get_rank()
        nprocs = MPI.COMM_WORLD.Get_size()
        m = MpiPartition(ngroups=1)

        self.assertEqual(m.ngroups, 1)

        self.assertEqual(m.rank_world, rank_world)
        self.assertEqual(m.rank_groups, rank_world)
        self.assertEqual(m.rank_leaders, 0 if rank_world == 0 else -1)

        self.assertEqual(m.nprocs_world, nprocs)
        self.assertEqual(m.nprocs_groups, nprocs)
        self.assertEqual(m.nprocs_leaders, 1 if rank_world == 0 else -1)

        self.assertEqual(m.proc0_world, rank_world == 0)
        self.assertEqual(m.proc0_groups, rank_world == 0)
        m.write()

    def test_ngroups_max(self):
        """
        Verify that all quantities make sense when ngroups >= nprocs_world
        and ngroups = None.
        """
        rank_world = MPI.COMM_WORLD.Get_rank()
        nprocs = MPI.COMM_WORLD.Get_size()

        for shift in range(-1, 3):
            if shift == -1:
                ngroups = None
            else:
                ngroups = nprocs + shift

            m = MpiPartition(ngroups=ngroups)

            self.assertEqual(m.ngroups, nprocs)

            self.assertEqual(m.rank_world, rank_world)
            self.assertEqual(m.rank_groups, 0)
            self.assertEqual(m.rank_leaders, rank_world)

            self.assertEqual(m.nprocs_world, nprocs)
            self.assertEqual(m.nprocs_groups, 1)
            self.assertEqual(m.nprocs_leaders, nprocs)

            self.assertEqual(m.proc0_world, rank_world == 0)
            self.assertTrue(m.proc0_groups)

    def test_ngroups_scan(self):
        """
        Verify that all quantities make sense when ngroups >= nprocs_world
        and ngroups = None.
        """
        rank_world = MPI.COMM_WORLD.Get_rank()
        nprocs = MPI.COMM_WORLD.Get_size()

        for ngroups in range(-1, nprocs+3):
            m = MpiPartition(ngroups=ngroups)

            self.assertGreaterEqual(m.ngroups, 1)
            self.assertLessEqual(m.ngroups, nprocs)

            self.assertEqual(m.rank_world, rank_world)
            self.assertGreaterEqual(m.rank_groups, 0)
            self.assertLess(m.rank_groups, nprocs)

            self.assertEqual(m.nprocs_world, nprocs)
            self.assertGreaterEqual(m.nprocs_groups, 1)
            self.assertLessEqual(m.nprocs_groups, nprocs)

            self.assertEqual(m.proc0_world, rank_world == 0)

            if m.proc0_groups:
                self.assertGreaterEqual(m.rank_leaders, 0)
                self.assertLessEqual(m.rank_leaders, nprocs)
                self.assertGreaterEqual(m.nprocs_leaders, 1)
                self.assertLessEqual(m.nprocs_leaders, nprocs)
            else:
                self.assertEqual(m.rank_leaders, -1)
                self.assertEqual(m.nprocs_leaders, -1)

            # The sizes of the worker groups should be relatively
            # even, with a difference of no more than 1 between the
            # largest and the smallest.
            if m.proc0_world:
                group_sizes = np.zeros(nprocs, dtype='i')
                group_sizes[0] = m.nprocs_groups
                for j in range(1, nprocs):
                    group_sizes[j] = m.comm_world.recv(tag=j)
                print('group_sizes:', group_sizes)
                self.assertLessEqual(np.max(group_sizes) - np.min(group_sizes), 1)
            else:
                m.comm_world.send(m.nprocs_groups, 0, tag=m.rank_world)
        m.write()

    @unittest.skip
    def test_fd_jac(self):
        """
        Test the parallel finite-difference Jacobian calculation.
        """
        for ngroups in range(1, 4):
            logger.debug('ngroups={}'.format(ngroups))
            mpi = MpiPartition(ngroups=ngroups)
            o = TestFunction1()
            logger.debug('About to do worker loop 1')
            jac, xs, evals = fd_jac_mpi(o, mpi, centered=False, eps=1e-7)
            jac_reference = np.array([[5.865176283537110e-01, -6.010834349701177e-01, 2.250910244305793e-01]])
            if mpi.proc0_world:
                np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            # o.set_dofs(np.array([1.2, 0.9, -0.4]))
            o.x = np.array([1.2, 0.9, -0.4])
            #jac = d.fd_jac(centered=False, eps=1e-7)
            #np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

            # Repeat with centered differences
            #o.set_dofs(np.array([1.2, 0.9, -0.4]))
            o.x = np.array([1.2, 0.9, -0.4])
            logger.debug('About to do worker loop 2')
            jac, xs, evals = fd_jac_mpi(o, mpi, centered=True, eps=1e-7)
            jac_reference = np.array([[5.865175337071982e-01, -6.010834789627051e-01, 2.250910093037906e-01]])
            if mpi.proc0_world:
                np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            # o.set_dofs(np.array([1.2, 0.9, -0.4]))
            o.x = np.array([1.2, 0.9, -0.4])
            # jac = o.fd_jac(centered=True, eps=1e-7)
            # np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

            # Now try a case with different nparams and nfuncs.
            o = TestFunction2()
            logger.debug('About to do worker loop 3')
            # jac, xs, evals = fd_jac_mpi(o, mpi, centered=False, eps=1e-7)
            # jac_reference = np.array([[8.657715439008840e-01, -8.872724499564555e-01],
            #                           [2.353411054922816e+00, -2.411856577788640e+00],
            #                           [6.397234502131255e+00, -6.556105911492693e+00],
            #                           [1.738948636642590e+01, -1.782134355643450e+01]])
            # if mpi.proc0_world:
            #     np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            # o.set_dofs(np.array([1.2, 0.9]))
            o.x = np.array([1.2, 0.9])
            # jac = d.fd_jac(centered=False, eps=1e-7)
            # np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

            # Repeat with centered differences
            # o.set_dofs(np.array([1.2, 0.9]))
            logger.debug('About to do worker loop 4')
            # jac, xs, evals = fd_jac_mpi(d, mpi, centered=True, eps=1e-7)
            # jac_reference = np.array([[8.657714037352271e-01, -8.872725151820582e-01],
            #                           [2.353410674116319e+00, -2.411856754314101e+00],
            #                           [6.397233469623842e+00, -6.556106388888594e+00],
            #                           [1.738948351093228e+01, -1.782134486205678e+01]])
            # if mpi.proc0_world:
            #     np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            # o.set_dofs(np.array([1.2, 0.9]))
            o.x = np.array([1.2, 0.9])
            # jac = d.fd_jac(centered=True, eps=1e-7)
            # np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

    def test_parallel_optimization(self):
        """
        Test a full least-squares optimization.
        """
        rank_world = MPI.COMM_WORLD.Get_rank()
        logger.info(f"rank world is {rank_world}")
        nprocs = MPI.COMM_WORLD.Get_size()
        for ngroups in range(1, 4):
            #for grad in [True, False]:
            mpi = MpiPartition(ngroups=ngroups)
            o = TestFunction3(mpi.comm_groups)
            term1 = (o.f0, 0, 1)
            term2 = (o.f1, 0, 1)
            prob = LeastSquaresProblem.from_tuples([term1, term2])
            least_squares_mpi_solve(prob, mpi)  # , grad=grad)
            self.assertAlmostEqual(prob.full_x[0], 1)
            self.assertAlmostEqual(prob.full_x[1], 1)



import unittest
import numpy as np
from scipy.optimize import minimize

from simsopt.objectives.functions import Identity, Rosenbrock, \
    Failer, RosenbrockWithFailures
from simsopt.objectives.least_squares import LeastSquaresProblem, \
    LeastSquaresTerm
from simsopt._core.optimizable import Target


class LeastSquaresTermTests(unittest.TestCase):

    def test_basic(self):
        """
        Test basic usage
        """
        iden = Identity()
        lst = LeastSquaresTerm(iden.J, 3, weight=0.1)
        self.assertEqual(lst.f_in, iden.J)
        self.assertEqual(lst.goal, 3)
        self.assertAlmostEqual(lst.weight, 0.1, places=13)

        lst = LeastSquaresTerm.from_sigma(iden.J, 3, sigma=0.1)
        self.assertEqual(lst.f_in, iden.J)
        self.assertEqual(lst.goal, 3)
        self.assertAlmostEqual(lst.weight, 100.0, places=13)

        iden.set_dofs([17])
        self.assertEqual(lst.f_in(), 17)
        correct_value = ((17 - 3) / 0.1) ** 2
        self.assertAlmostEqual(lst.f_out(), correct_value, places=11)

    def test_supply_object(self):
        """
        Test that we can supply an object with a J function instead of a function.
        """
        iden = Identity()
        # Note here we supply iden instead of iden.J
        lst = LeastSquaresTerm.from_sigma(iden, 3, sigma=0.1)
        self.assertEqual(lst.f_in, iden.J)
        self.assertEqual(lst.goal, 3)
        self.assertAlmostEqual(lst.weight, 100, places=13)

        iden.set_dofs([17])
        self.assertEqual(lst.f_in(), 17)
        correct_value = ((17 - 3) / 0.1) ** 2
        self.assertAlmostEqual(lst.f_out(), correct_value, places=11)

    def test_supply_property(self):
        """
        Test that we can supply a property instead of a function.
        """
        iden = Identity()
        lst = LeastSquaresTerm.from_sigma(Target(iden, 'f'), 3, sigma=0.1)
        self.assertEqual(lst.goal, 3)
        self.assertAlmostEqual(lst.weight, 100, places=13)

        iden.set_dofs([17])
        self.assertEqual(lst.f_in(), 17)
        correct_value = ((17 - 3) / 0.1) ** 2
        self.assertAlmostEqual(lst.f_out(), correct_value, places=11)

    def test_supply_attribute(self):
        """
        Test that we can supply an attribute instead of a function.
        """
        iden = Identity()
        lst = LeastSquaresTerm(Target(iden, 'x'), 3, weight=0.1)
        self.assertEqual(lst.goal, 3)
        self.assertAlmostEqual(lst.weight, 0.1, places=13)

        iden.set_dofs([17])
        self.assertEqual(lst.f_in(), 17)
        correct_value = 0.1 * ((17 - 3) ** 2)
        self.assertAlmostEqual(lst.f_out(), correct_value, places=13)

    def test_exceptions(self):
        """
        Test that exceptions are thrown when invalid inputs are
        provided.
        """
        # First argument must be callable
        with self.assertRaises(TypeError):
            lst = LeastSquaresTerm(2, 3, 0.1)

        # Second and third arguments must be real numbers
        iden = Identity()
        #with self.assertRaises(TypeError):
        #    lst = LeastSquaresTerm(iden.J, "hello", sigma=0.1)
        with self.assertRaises(TypeError):
            lst = LeastSquaresTerm.from_sigma(iden.J, 3, sigma=iden)
        #with self.assertRaises(TypeError):
        #    lst = LeastSquaresTerm(iden.J, "hello", weight=0.1)
        with self.assertRaises(TypeError):
            lst = LeastSquaresTerm(iden.J, 3, weight=iden)

        # sigma cannot be zero
        with self.assertRaises(ValueError):
            lst = LeastSquaresTerm.from_sigma(iden.J, 3, sigma=0)
        with self.assertRaises(ValueError):
            lst = LeastSquaresTerm.from_sigma(iden.J, 3, sigma=0.0)

        # Cannot specify both weight and sigma
        with self.assertRaises(TypeError):
            lst = LeastSquaresTerm(iden.J, 3, sigma=1.2, weight=3.4)
        # Must specify either weight or sigma
        with self.assertRaises(TypeError):
            lst = LeastSquaresTerm(iden.J, 3)

        # Weight cannot be negative
        with self.assertRaises(ValueError):
            lst = LeastSquaresTerm(iden.J, 3, weight=-1.0)


class LeastSquaresProblemTests(unittest.TestCase):

    def test_supply_LeastSquaresTerm(self):
        """
        Test basic usage
        """
        # Objective function f(x) = ((x - 3) / 2) ** 2
        iden1 = Identity()
        term1 = LeastSquaresTerm.from_sigma(iden1.J, 3, sigma=2)
        prob = LeastSquaresProblem([term1])
        self.assertAlmostEqual(prob.objective(), 2.25)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertEqual(len(prob.dofs.f()), 1)
        self.assertAlmostEqual(prob.dofs.f()[0], 0)
        self.assertEqual(len(prob.f()), 1)
        self.assertAlmostEqual(prob.f()[0], -1.5)
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 2.25)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 2.25)
        iden1.set_dofs([10])
        self.assertAlmostEqual(prob.objective(), 12.25)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 12.25)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 12.25)
        self.assertAlmostEqual(prob.objective([0]), 2.25)
        self.assertAlmostEqual(prob.objective([10]), 12.25)
        self.assertEqual(prob.dofs.all_owners, [iden1])
        self.assertEqual(prob.dofs.dof_owners, [iden1])

        # Objective function
        # f(x,y) = ((x - 3) / 2) ** 2 + ((y + 4) / 5) ** 2
        iden2 = Identity()
        term2 = LeastSquaresTerm.from_sigma(iden2.J, -4, sigma=5)
        prob = LeastSquaresProblem([term1, term2])
        self.assertAlmostEqual(prob.objective(), 12.89)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertEqual(len(prob.f()), 2)
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 12.89)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 12.89)
        iden1.set_dofs([5])
        iden2.set_dofs([-7])
        self.assertAlmostEqual(prob.objective(), 1.36)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertEqual(len(prob.f()), 2)
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 1.36)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 1.36)
        self.assertAlmostEqual(prob.objective([10, 0]), 12.89)
        self.assertAlmostEqual(prob.objective([5, -7]), 1.36)
        self.assertEqual(prob.dofs.dof_owners, [iden1, iden2])
        self.assertEqual(prob.dofs.all_owners, [iden1, iden2])

    def test_supply_tuples(self):
        """
        Test basic usage
        """
        # Objective function f(x) = ((x - 3) / 2) ** 2
        iden1 = Identity()
        term1 = (iden1.J, 3, 0.25)
        prob = LeastSquaresProblem([term1])
        self.assertAlmostEqual(prob.objective(), 2.25)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertEqual(len(prob.dofs.f()), 1)
        self.assertAlmostEqual(prob.dofs.f()[0], 0)
        self.assertEqual(len(prob.f()), 1)
        self.assertAlmostEqual(prob.f()[0], -1.5)
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 2.25)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 2.25)
        iden1.set_dofs([10])
        self.assertAlmostEqual(prob.objective(), 12.25)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 12.25)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 12.25)
        self.assertAlmostEqual(prob.objective([0]), 2.25)
        self.assertAlmostEqual(prob.objective([10]), 12.25)
        self.assertEqual(prob.dofs.all_owners, [iden1])
        self.assertEqual(prob.dofs.dof_owners, [iden1])

        # Objective function
        # f(x,y) = ((x - 3) / 2) ** 2 + ((y + 4) / 5) ** 2
        iden2 = Identity()
        term2 = (iden2.J, -4, 0.04)
        prob = LeastSquaresProblem([term1, term2])
        self.assertAlmostEqual(prob.objective(), 12.89)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertEqual(len(prob.f()), 2)
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 12.89)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 12.89)
        iden1.set_dofs([5])
        iden2.set_dofs([-7])
        self.assertAlmostEqual(prob.objective(), 1.36)
        self.assertAlmostEqual(prob.objective(), sum(t.f_out() for t in prob.terms))
        self.assertEqual(len(prob.f()), 2)
        self.assertAlmostEqual(prob.objective_from_shifted_f(prob.f()), 1.36)
        self.assertAlmostEqual(prob.objective_from_unshifted_f(prob.dofs.f()), 1.36)
        self.assertAlmostEqual(prob.objective([10, 0]), 12.89)
        self.assertAlmostEqual(prob.objective([5, -7]), 1.36)
        self.assertEqual(prob.dofs.dof_owners, [iden1, iden2])
        self.assertEqual(prob.dofs.all_owners, [iden1, iden2])

    def test_exceptions(self):
        """
        Verify that exceptions are raised when invalid inputs are
        provided.
        """
        # Argument must be a list in which each element is a
        # LeastSquaresTerm or a 3- or 4-element tuple/list.
        with self.assertRaises(TypeError):
            prob = LeastSquaresProblem(7)
        with self.assertRaises(ValueError):
            prob = LeastSquaresProblem([])
        with self.assertRaises(TypeError):
            prob = LeastSquaresProblem([7, 1])

    def test_failures(self):
        """
        Verify that the expected residuals are returned in cases where the
        objective function evaluations fail.
        """
        o1 = Failer()
        r1 = Rosenbrock()
        fail_val = 1.0e6
        prob1 = LeastSquaresProblem([(r1.terms, 0, 1), (o1, 0, 1)],
                                    fail=fail_val)
        # First evaluation should not fail.
        f = prob1.f()
        print(f)
        np.testing.assert_allclose(f, [-1, 0, 1, 1, 1])
        # Second evaluation should fail.
        f = prob1.f()
        print(f)
        np.testing.assert_allclose(f, np.full(5, fail_val))
        # Third evaluation should not fail.
        f = prob1.f()
        print(f)
        np.testing.assert_allclose(f, [-1, 0, 1, 1, 1])

    def test_outside_optimizer(self):
        """
        Verify that a least-squares problem can be passed to an outside
        optimization package to be solved, even when failures occur
        during evaluations.
        """
        ros = RosenbrockWithFailures(fail_interval=20)
        fail_val = 1.0e2
        prob = LeastSquaresProblem([(ros.terms, 0, 1)], fail=fail_val)
        # Just call the bare scipy.optimize.minimize function. This is
        # not really an "outside" package, but it is similar to how we
        # might want to call outside optimization libraries that are
        # completely separate from simsopt.
        result = minimize(prob.objective, prob.x)
        # Need a large tolerance, since the failures confuse the
        # optimizer
        np.testing.assert_allclose(result.x, [1, 1], atol=1e-2)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
from simsopt.field.coil import coils_via_symmetries
from simsopt.geo.boozersurface import BoozerSurface
from simsopt.field.biotsavart import BiotSavart
from simsopt.geo.surfaceobjectives import ToroidalFlux
from simsopt.geo.surfaceobjectives import Area
from simsopt.util.zoo import get_ncsx_data
from .surface_test_helpers import get_surface, get_exact_surface


surfacetypes_list = ["SurfaceXYZFourier", "SurfaceXYZTensorFourier"]
stellsym_list = [True, False]


class BoozerSurfaceTests(unittest.TestCase):
    def test_residual(self):
        """
        This test loads a SurfaceXYZFourier that interpolates the xyz coordinates
        of a surface in the NCSX configuration that was computed on a previous
        branch of pyplasmaopt.  Here, we verify that the  Boozer residual at these 
        interpolation points is small.
        """

        s = get_exact_surface()
        curves, currents, ma = get_ncsx_data()
        coils = coils_via_symmetries(curves, currents, 3, True)
        bs = BiotSavart(coils)
        bs_tf = BiotSavart(coils)

        weight = 1.
        tf = ToroidalFlux(s, bs_tf)

        # these data are obtained from `boozer` branch of pyplamsaopt
        tf_target = 0.41431152
        iota = -0.44856192

        boozer_surface = BoozerSurface(bs, s, tf, tf_target)
        x = np.concatenate((s.get_dofs(), [iota]))
        r0 = boozer_surface.boozer_penalty_constraints(
            x, derivatives=0, constraint_weight=weight, optimize_G=False, scalarize=False)
        # the residual should be close to zero for all entries apart from the y
        # and z coordinate at phi=0 and theta=0 (and the corresponding rotations)
        ignores_idxs = np.zeros_like(r0)
        ignores_idxs[[1, 2, 693, 694, 695, 1386, 1387, 1388, -2, -1]] = 1
        assert np.max(np.abs(r0[ignores_idxs < 0.5])) < 1e-8
        assert np.max(np.abs(r0[-2:])) < 1e-6

    def test_boozer_penalty_constraints_gradient(self):
        """
        Taylor test to verify the gradient of the scalarized constrained optimization problem's
        objective.
        """
        for surfacetype in surfacetypes_list:
            for stellsym in stellsym_list:
                for optimize_G in [True, False]:
                    with self.subTest(surfacetype=surfacetype, stellsym=stellsym, optimize_G=optimize_G):
                        self.subtest_boozer_penalty_constraints_gradient(surfacetype, stellsym, optimize_G)

    def test_boozer_penalty_constraints_hessian(self):
        """
        Taylor test to verify the Hessian of the scalarized constrained optimization problem's
        objective.
        """
        for surfacetype in surfacetypes_list:
            for stellsym in stellsym_list:
                for optimize_G in [True, False]:
                    with self.subTest(surfacetype=surfacetype, stellsym=stellsym, optimize_G=optimize_G):
                        self.subtest_boozer_penalty_constraints_hessian(surfacetype, stellsym, optimize_G)

    def subtest_boozer_penalty_constraints_gradient(self, surfacetype, stellsym, optimize_G=False):
        np.random.seed(1)
        curves, currents, ma = get_ncsx_data()
        coils = coils_via_symmetries(curves, currents, 3, True)
        bs = BiotSavart(coils)
        bs_tf = BiotSavart(coils)
        current_sum = sum(abs(c.current.get_value()) for c in coils)

        s = get_surface(surfacetype, stellsym)
        s.fit_to_curve(ma, 0.1)

        weight = 11.1232

        tf = ToroidalFlux(s, bs_tf)

        tf_target = 0.1
        boozer_surface = BoozerSurface(bs, s, tf, tf_target)

        iota = -0.3
        x = np.concatenate((s.get_dofs(), [iota]))
        if optimize_G:
            x = np.concatenate((x, [2.*np.pi*current_sum*(4*np.pi*10**(-7)/(2 * np.pi))]))
        f0, J0 = boozer_surface.boozer_penalty_constraints(
            x, derivatives=1, constraint_weight=weight, optimize_G=optimize_G)

        h = np.random.uniform(size=x.shape)-0.5
        Jex = J0@h

        err_old = 1e9
        epsilons = np.power(2., -np.asarray(range(7, 20)))
        print("################################################################################")
        for eps in epsilons:
            f1 = boozer_surface.boozer_penalty_constraints(
                x + eps*h, derivatives=0, constraint_weight=weight, optimize_G=optimize_G)
            Jfd = (f1-f0)/eps
            err = np.linalg.norm(Jfd-Jex)/np.linalg.norm(Jex)
            print(err/err_old, f0, f1)
            assert err < err_old * 0.55
            err_old = err
        print("################################################################################")

    def subtest_boozer_penalty_constraints_hessian(self, surfacetype, stellsym, optimize_G=False):
        np.random.seed(1)
        curves, currents, ma = get_ncsx_data()
        coils = coils_via_symmetries(curves, currents, 3, True)
        bs = BiotSavart(coils)
        bs_tf = BiotSavart(coils)
        current_sum = sum(abs(c.current.get_value()) for c in coils)

        s = get_surface(surfacetype, stellsym)
        s.fit_to_curve(ma, 0.1)

        tf = ToroidalFlux(s, bs_tf)

        tf_target = 0.1
        boozer_surface = BoozerSurface(bs, s, tf, tf_target)

        iota = -0.3
        x = np.concatenate((s.get_dofs(), [iota]))
        if optimize_G:
            x = np.concatenate((x, [2.*np.pi*current_sum*(4*np.pi*10**(-7)/(2 * np.pi))]))
        f0, J0, H0 = boozer_surface.boozer_penalty_constraints(x, derivatives=2, optimize_G=optimize_G)

        h1 = np.random.uniform(size=x.shape)-0.5
        h2 = np.random.uniform(size=x.shape)-0.5
        d2f = h1@H0@h2

        err_old = 1e9
        epsilons = np.power(2., -np.asarray(range(10, 20)))
        print("################################################################################")
        for eps in epsilons:
            fp, Jp = boozer_surface.boozer_penalty_constraints(x + eps*h1, derivatives=1, optimize_G=optimize_G)
            d2f_fd = (Jp@h2-J0@h2)/eps
            err = np.abs(d2f_fd-d2f)/np.abs(d2f)
            print(err/err_old)
            assert err < err_old * 0.55
            err_old = err

    def test_boozer_constrained_jacobian(self):
        """
        Taylor test to verify the Jacobian of the first order optimality conditions of the exactly
        constrained optimization problem.
        """
        for surfacetype in surfacetypes_list:
            for stellsym in stellsym_list:
                for optimize_G in [True, False]:
                    with self.subTest(surfacetype=surfacetype, stellsym=stellsym, optimize_G=optimize_G):
                        self.subtest_boozer_constrained_jacobian(surfacetype, stellsym, optimize_G)

    def subtest_boozer_constrained_jacobian(self, surfacetype, stellsym, optimize_G=False):
        np.random.seed(1)
        curves, currents, ma = get_ncsx_data()
        coils = coils_via_symmetries(curves, currents, 3, True)
        bs = BiotSavart(coils)
        bs_tf = BiotSavart(coils)
        current_sum = sum(abs(c.current.get_value()) for c in coils)

        s = get_surface(surfacetype, stellsym)
        s.fit_to_curve(ma, 0.1)

        tf = ToroidalFlux(s, bs_tf)

        tf_target = 0.1
        boozer_surface = BoozerSurface(bs, s, tf, tf_target)

        iota = -0.3
        lm = [0., 0.]
        x = np.concatenate((s.get_dofs(), [iota]))
        if optimize_G:
            x = np.concatenate((x, [2.*np.pi*current_sum*(4*np.pi*10**(-7)/(2 * np.pi))]))
        xl = np.concatenate((x, lm))
        res0, dres0 = boozer_surface.boozer_exact_constraints(xl, derivatives=1, optimize_G=optimize_G)

        h = np.random.uniform(size=xl.shape)-0.5
        dres_exact = dres0@h

        err_old = 1e9
        epsilons = np.power(2., -np.asarray(range(7, 20)))
        print("################################################################################")
        for eps in epsilons:
            res1 = boozer_surface.boozer_exact_constraints(xl + eps*h, derivatives=0, optimize_G=optimize_G)
            dres_fd = (res1-res0)/eps
            err = np.linalg.norm(dres_fd-dres_exact)
            print(err/err_old)
            assert err < err_old * 0.55
            err_old = err
        print("################################################################################")

    def test_boozer_surface_optimisation_convergence(self):
        """
        Test to verify the various optimization algorithms that compute
        the Boozer angles on a surface.
        """

        configs = [
            ("SurfaceXYZTensorFourier", True, True, 'residual_exact'),  # noqa
            ("SurfaceXYZTensorFourier", True, True, 'newton_exact'),  # noqa
            ("SurfaceXYZTensorFourier", True, True, 'newton'),  # noqa
            ("SurfaceXYZTensorFourier", False, True, 'ls'),  # noqa
            ("SurfaceXYZFourier", True, False, 'ls'),  # noqa
        ]
        for surfacetype, stellsym, optimize_G, second_stage in configs:
            with self.subTest(
                surfacetype=surfacetype, stellsym=stellsym,
                    optimize_G=optimize_G, second_stage=second_stage):
                self.subtest_boozer_surface_optimisation_convergence(surfacetype, stellsym, optimize_G, second_stage)

    def subtest_boozer_surface_optimisation_convergence(self, surfacetype, stellsym, optimize_G, second_stage):
        curves, currents, ma = get_ncsx_data()

        if stellsym:
            coils = coils_via_symmetries(curves, currents, 3, True)
        else:
            # Create a stellarator that still has rotational symmetry but
            # doesn't have stellarator symmetry. We do this by first applying
            # stellarator symmetry, then breaking this slightly, and then
            # applying rotational symmetry
            from simsopt.geo.curve import RotatedCurve
            curves_flipped = [RotatedCurve(c, 0, True) for c in curves]
            currents_flipped = [-cur for cur in currents]
            for c in curves_flipped:
                c.rotmat += 0.001*np.random.uniform(low=-1., high=1., size=c.rotmat.shape)
                c.rotmatT = c.rotmat.T
            coils = coils_via_symmetries(curves + curves_flipped, currents + currents_flipped, 3, False)
        current_sum = sum(abs(c.current.get_value()) for c in coils)

        bs = BiotSavart(coils)

        s = get_surface(surfacetype, stellsym)
        s.fit_to_curve(ma, 0.1)
        iota = -0.3

        ar = Area(s)
        ar_target = ar.J()
        boozer_surface = BoozerSurface(bs, s, ar, ar_target)

        if optimize_G:
            G = 2.*np.pi*current_sum*(4*np.pi*10**(-7)/(2 * np.pi))
        else:
            G = None

        # compute surface first using LBFGS exact and an area constraint
        res = boozer_surface.minimize_boozer_penalty_constraints_LBFGS(
            tol=1e-9, maxiter=500, constraint_weight=100., iota=iota, G=G)
        print('Residual norm after LBFGS', np.sqrt(2*res['fun']))
        if second_stage == 'ls':
            res = boozer_surface.minimize_boozer_penalty_constraints_ls(
                tol=1e-11, maxiter=100, constraint_weight=100., iota=res['iota'], G=res['G'])
        elif second_stage == 'newton':
            res = boozer_surface.minimize_boozer_penalty_constraints_newton(
                tol=1e-9, maxiter=10, constraint_weight=100., iota=res['iota'], G=res['G'], stab=1e-4)
        elif second_stage == 'newton_exact':
            res = boozer_surface.minimize_boozer_exact_constraints_newton(
                tol=1e-9, maxiter=10, iota=res['iota'], G=res['G'])
        elif second_stage == 'residual_exact':
            res = boozer_surface.solve_residual_equation_exactly_newton(
                tol=1e-12, maxiter=10, iota=res['iota'], G=res['G'])

        print('Residual norm after second stage', np.linalg.norm(res['residual']))
        assert res['success']
        # For the stellsym case we have z(0, 0) = y(0, 0) = 0. For the not
        # stellsym case, we enforce z(0, 0) = 0, but expect y(0, 0) \neq 0
        gammazero = s.gamma()[0, 0, :]
        assert np.abs(gammazero[2]) < 1e-10
        if stellsym:
            assert np.abs(gammazero[1]) < 1e-10
        else:
            assert np.abs(gammazero[1]) > 1e-6

        if surfacetype == 'SurfaceXYZTensorFourier':
            assert np.linalg.norm(res['residual']) < 1e-9

        print(ar_target, ar.J())
        print(res['residual'][-10:])
        if surfacetype == 'SurfaceXYZTensorFourier' or second_stage == 'newton_exact':
            assert np.abs(ar_target - ar.J()) < 1e-9
        else:
            assert np.abs(ar_target - ar.J()) < 1e-4


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import logging
import numpy as np

from simsopt.util.mpi import MpiPartition, log
from simsopt.mhd import Vmec, Spec
from simsopt import LeastSquaresProblem
from simsopt.solve.mpi import least_squares_mpi_solve
import os

"""
This script implements the "2DOF_vmecOnly_targetIotaAndVolume" example from
https://github.com/landreman/stellopt_scenarios

This optimization problem has two independent variables, representing
the helical shape of the magnetic axis. The problem also has two
objectives: the plasma volume and the rotational transform on the
magnetic axis.

The resolution in this example (i.e. ns, mpol, and ntor) is somewhat
lower than in the stellopt_scenarios version of the example, just so
this example runs fast.

Details of the optimum and a plot of the objective function landscape
can be found here:
https://github.com/landreman/stellopt_scenarios/tree/master/2DOF_vmecOnly_targetIotaAndVolume
"""

# This next line turns on detailed logging. It can be commented out if
# you do not want such verbose output.
log(logging.INFO)

mpi = MpiPartition()

# Initialize VMEC from an input file:
vmec = Vmec(os.path.join(os.path.dirname(__file__), 'inputs', 'input.2DOF_vmecOnly_targetIotaAndVolume'), mpi=mpi)
surf = vmec.boundary

# Initialize SPEC from an input file:
spec = Spec(os.path.join(os.path.dirname(__file__), 'inputs', '2DOF_targetIotaAndVolume.sp'), mpi=mpi)

# Set the SPEC boundary to be the same object as the VMEC boundary!
spec.boundary = surf

# VMEC parameters are all fixed by default, while surface parameters are all non-fixed by default.
# You can choose which parameters are optimized by setting their 'fixed' attributes.
surf.all_fixed()
surf.set_fixed('rc(1,1)', False)
surf.set_fixed('zs(1,1)', False)

# Each Target is then equipped with a shift and weight, to become a
# term in a least-squares objective function.  A list of terms are
# combined to form a nonlinear-least-squares problem.
desired_volume = 0.15
volume_weight = 1
term1 = (spec.volume, desired_volume, volume_weight)

desired_iota = 0.41
iota_weight = 1
term2 = (vmec.iota_axis, desired_iota, iota_weight)

prob = LeastSquaresProblem([term1, term2])

# Solve the minimization problem:
least_squares_mpi_solve(prob, mpi=mpi, grad=True)

# Evaluate quantities on all processes, in case communication is
# required:
final_objective = prob.objective()
vmec_volume = vmec.volume()
spec_volume = spec.volume()
surf_volume = surf.volume()
vmec_iota = vmec.iota_axis()
spec_iota = spec.iota()

if mpi.proc0_world:
    logging.info("At the optimum,")
    logging.info(f" objective function = {final_objective}")
    logging.info(f" rc(m=1,n=1) = {surf.get_rc(1, 1)}")
    logging.info(f" zs(m=1,n=1) = {surf.get_zs(1, 1)}")
    logging.info(f" volume, according to VMEC    = {vmec_volume}")
    logging.info(f" volume, according to SPEC    = {spec_volume}")
    logging.info(f" volume, according to Surface = {surf_volume}")
    logging.info(f" iota on axis, from VMEC       = {vmec_iota}")
    logging.info(f" iota at mid-radius, from SPEC = {spec_iota}")

    assert np.abs(surf.get_rc(1, 1) - 0.0313066948) < 1.0e-3
    assert np.abs(surf.get_zs(1, 1) - (-0.031232391)) < 1.0e-3
    assert np.abs(spec_volume - 0.178091) < 1.0e-3
    assert np.abs(vmec_volume - 0.178091) < 1.0e-3
    assert np.abs(surf_volume - 0.178091) < 1.0e-3
    assert np.abs(vmec_iota - 0.4114567) < 1.0e-4
    assert final_objective < 1.0e-2



from simsopt.field.coil import coils_via_symmetries
from simsopt.field.biotsavart import BiotSavart
from simsopt.geo.curvexyzfourier import CurveXYZFourier
from simsopt.util.zoo import get_ncsx_data
from simsopt.field.tracing import trace_particles_starting_on_curve, SurfaceClassifier, \
    particles_to_vtk, LevelsetStoppingCriterion, compute_gc_radius, gc_to_fullorbit_initial_guesses, \
    IterationStoppingCriterion, trace_particles_starting_on_surface
from simsopt.geo.surfacerzfourier import SurfaceRZFourier
from simsopt.field.magneticfieldclasses import InterpolatedField, UniformInterpolationRule, ToroidalField, PoloidalField
from simsopt.util.constants import PROTON_MASS, ELEMENTARY_CHARGE, ONE_EV
import simsoptpp as sopp
import numpy as np
import unittest
import logging
logging.basicConfig()
try:
    import pyevtk
    with_evtk = True
except ImportError:
    with_evtk = False


def validate_phi_hits(phi_hits, bfield, nphis):
    """
    After hitting one of the phi planes, three things can happen:
    A: We can hit the same phi plane again (i.e. the particle bounces)
    B: We can hit one of the stopping criteria
    C: We can hit the next phi plane. Note that the definition of 'next'
       depends on the direction the particle is going. By looking at the
       direction of the B field and the tangential velocity, we can figure out
       if the particle was going clockwise or anti clockwise, and hence we know
       which is the 'next' phi plane.
    """
    for i in range(len(phi_hits)-1):
        this_idx = int(phi_hits[i][1])
        this_B = bfield.set_points(np.asarray(
            phi_hits[i][2:5]).reshape((1, 3))).B_cyl()
        vtang = phi_hits[i][5]
        forward = np.sign(vtang * this_B[0, 1]) > 0
        next_idx = int(phi_hits[i+1][1])
        hit_same_phi = next_idx == this_idx
        hit_stopping_criteria = next_idx < 0
        hit_next_phi = next_idx == (this_idx + 1) % nphis
        hit_previous_phi = next_idx == (this_idx - 1) % nphis
        if forward:
            if not (hit_next_phi or hit_same_phi or hit_stopping_criteria):
                return False
        else:
            if not (hit_previous_phi or hit_same_phi or hit_stopping_criteria):
                return False
    return True


class ParticleTracingTesting(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(ParticleTracingTesting, self).__init__(*args, **kwargs)
        logger = logging.getLogger('simsopt.field.tracing')
        logger.setLevel(1)
        curves, currents, ma = get_ncsx_data()
        nfp = 3
        coils = coils_via_symmetries(curves, currents, nfp, True)
        bs = BiotSavart(coils)
        n = 16

        rrange = (1.1, 1.8, n)
        phirange = (0, 2*np.pi/3, n*2)
        zrange = (0, 0.3, n//2)
        bsh = InterpolatedField(
            bs, UniformInterpolationRule(5),
            rrange, phirange, zrange, True, nfp=3, stellsym=True
        )
        print(bsh.estimate_error_B(1000))
        print(bsh.estimate_error_GradAbsB(1000))
        # trick to clear the cache in the biot savart class to reduce memory usage
        bs.set_points(np.asarray([[0., 0., 0.]])).GradAbsB()
        self.bsh = bsh
        self.ma = ma
        if with_evtk:
            bsh.to_vtk('/tmp/bfield')

    def test_guidingcenter_vs_fullorbit(self):
        bsh = self.bsh
        ma = self.ma
        nparticles = 2
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        Ekin = 1000*ONE_EV
        np.random.seed(1)
        umin = 0.25
        umax = 0.75
        tmax = 2e-5
        with self.assertRaises(RuntimeError):
            gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
                ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
                Ekin=Ekin, umin=umin, umax=umax,
                phis=[], mode='gc')

        gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=umin, umax=umax,
            phis=[], mode='gc_vac')
        fo_tys, fo_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=umin, umax=umax,
            phis=[], mode='full')
        if with_evtk:
            particles_to_vtk(gc_tys, '/tmp/particles_gc')
            particles_to_vtk(fo_tys, '/tmp/particles_fo')

        # pick 100 random points on each trace, and ensure that the guiding
        # center and the full orbit simulation are close to each other

        N = 100
        for i in range(nparticles):
            gc_ty = gc_tys[i]
            fo_ty = fo_tys[i]
            idxs = np.random.randint(0, gc_ty.shape[0], size=(N, ))
            ts = gc_ty[idxs, 0]
            gc_xyzs = gc_ty[idxs, 1:4]
            bsh.set_points(gc_xyzs)
            Bs = bsh.B()
            AbsBs = bsh.AbsB()
            for j in range(N):
                jdx = np.argmin(np.abs(ts[j]-fo_ty[:, 0]))
                v = fo_ty[jdx, 4:]
                Bunit = Bs[j, :]/AbsBs[j, 0]
                vperp = np.linalg.norm(v - np.sum(v*Bunit)*Bunit)
                r = compute_gc_radius(m, vperp, q, AbsBs[j, 0])
                dist = np.linalg.norm(gc_xyzs[j, :] - fo_ty[jdx, 1:4])
                assert dist < 8*r

    def test_guidingcenterphihits(self):
        bsh = self.bsh
        ma = self.ma
        nparticles = 2
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        Ekin = 9000 * ONE_EV
        nphis = 10
        phis = np.linspace(0, 2*np.pi, nphis, endpoint=False)
        mpol = 5
        ntor = 5
        nfp = 3
        s = SurfaceRZFourier(
            mpol=mpol, ntor=ntor, stellsym=True, nfp=nfp,
            quadpoints_phi=np.linspace(0, 1, nfp*2*ntor+1, endpoint=False),
            quadpoints_theta=np.linspace(0, 1, 2*mpol+1, endpoint=False))
        s.fit_to_curve(ma, 0.10, flip_theta=False)
        sc = SurfaceClassifier(s, h=0.1, p=2)
        if with_evtk:
            sc.to_vtk('/tmp/classifier')
        # check that the axis is classified as inside the domain
        assert sc.evaluate(ma.gamma()[:1, :]) > 0
        assert sc.evaluate(2*ma.gamma()[:1, :]) < 0
        np.random.seed(1)
        gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=1e-4, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=-0.1, umax=+0.1, phis=phis, mode='gc_vac',
            stopping_criteria=[LevelsetStoppingCriterion(sc)], forget_exact_path=True)

        for i in range(nparticles):
            assert gc_tys[i].shape[0] == 2
            assert validate_phi_hits(gc_phi_hits[i], bsh, nphis)

    def test_gc_to_full(self):
        N = 100
        etas = np.linspace(0, 2*np.pi, N, endpoint=False)
        ma = self.ma
        bsh = self.bsh
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        Ekin = 9000*ONE_EV
        vtotal = np.sqrt(2*Ekin/m)  # Ekin = 0.5 * m * v^2 <=> v = sqrt(2*Ekin/m)
        xyz_gc = ma.gamma()[:10, :]
        np.random.seed(1)
        vtangs = np.random.uniform(size=(xyz_gc.shape[0], )) * vtotal
        xyz, vxyz, _ = gc_to_fullorbit_initial_guesses(bsh, xyz_gc, vtangs, vtotal, m, q, eta=etas[0])
        for eta in etas[1:]:
            tmp = gc_to_fullorbit_initial_guesses(bsh, xyz_gc, vtangs, vtotal, m, q, eta=eta)
            xyz += tmp[0]
            vxyz += tmp[1]
            radius = tmp[2]
            assert np.allclose(np.abs(np.linalg.norm(xyz_gc-tmp[0], axis=1) - radius), 0)
            assert np.all(np.abs(np.linalg.norm(tmp[1], axis=1) - vtotal)/vtotal < 1e-10)
        xyz *= 1/N
        vxyz *= 1/N
        assert np.linalg.norm(xyz - xyz_gc) < 1e-5
        B = bsh.set_points(xyz_gc).B()
        B *= 1./bsh.AbsB()
        assert np.linalg.norm((vxyz - B * vtangs[:, None])/vtotal) < 1e-5

    def test_energy_conservation(self):
        # Test conservation of Energy = m v^2/2=(m/2) (vparallel^2+ 2muB)
        # where mu=v_perp^2/(2B) is the adiabatic invariant. In the
        # the presence of a strong magnetic field, mu should also be
        # conserved. Energy is conserved regardless of the magnitude
        # of the magnetic field.
        bsh = self.bsh
        ma = self.ma
        nparticles = 1
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        tmax = 1e-5
        Ekin = 100*ONE_EV
        np.random.seed(1)

        gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=-0.5, umax=-0.25,  # pitch angle so that we have both par and perp contribution
            phis=[], mode='gc_vac', tol=1e-11)
        fo_tys, fo_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=-0.5, umax=-0.25,
            phis=[], mode='full', tol=1e-11)
        if with_evtk:
            particles_to_vtk(gc_tys, '/tmp/particles_gc')
            particles_to_vtk(fo_tys, '/tmp/particles_fo')

        # pick 100 random points on each trace, and ensure that
        # the energy is being conserved both in the guiding center
        # and full orbit cases up to some precision

        N = 100
        max_energy_gc_error = np.array([])
        max_energy_fo_error = np.array([])
        max_mu_gc_error = np.array([])
        max_mu_fo_error = np.array([])
        for i in range(nparticles):
            fo_ty = fo_tys[i]
            gc_ty = gc_tys[i]
            idxs = np.random.randint(0, gc_ty.shape[0], size=(N, ))
            gc_xyzs = gc_ty[idxs, 1:4]
            fo_xyzs = fo_ty[idxs, 1:4]
            bsh.set_points(gc_xyzs)
            AbsBs_gc = bsh.AbsB()
            bsh.set_points(fo_xyzs)
            AbsBs = bsh.AbsB()
            Bs = bsh.B()

            energy_fo = np.array([])
            energy_gc = np.array([])
            mu_fo = np.array([])
            mu_gc = np.array([])
            vParInitial = gc_ty[0, 4]
            muInitial = Ekin/(m*AbsBs_gc[0, 0])-vParInitial**2/(2*AbsBs_gc[0, 0])
            for j in range(N):
                v_fo = fo_ty[idxs[j], 4:]
                v_gc = gc_ty[idxs[j], 4]
                Bunit = Bs[j, :]/AbsBs[j, 0]
                vperp = np.linalg.norm(v_fo - np.sum(v_fo*Bunit)*Bunit)
                energy_fo = np.append(energy_fo, (m/2)*(v_fo[[0]]**2+v_fo[[1]]**2+v_fo[[2]]**2))
                energy_gc = np.append(energy_gc, (m/2)*(v_gc**2+2*muInitial*AbsBs_gc[j, 0]))
                mu_fo = np.append(mu_fo, vperp**2/(2*AbsBs[j, 0]))
                mu_gc = np.append(mu_gc, Ekin/(m*AbsBs_gc[j, 0])-v_gc**2/(2*AbsBs_gc[j, 0]))
            energy_fo_error = np.log10(np.abs(energy_fo-energy_fo[0]+1e-30)/energy_fo[0])
            energy_gc_error = np.log10(np.abs(energy_gc-energy_gc[0]+1e-30)/energy_gc[0])
            mu_fo_error = np.log10(np.abs(mu_fo-mu_fo[0]+1e-30)/mu_fo[0])
            mu_gc_error = np.log10(np.abs(mu_gc-mu_gc[0]+1e-30)/mu_gc[0])
            max_energy_fo_error = np.append(max_energy_fo_error, max(energy_fo_error[3::]))
            max_energy_gc_error = np.append(max_energy_gc_error, max(energy_gc_error[3::]))
            max_mu_fo_error = np.append(max_mu_fo_error, max(mu_fo_error[3::]))
            max_mu_gc_error = np.append(max_mu_gc_error, max(mu_gc_error[3::]))
        print("Max Energy Error for full orbit     = ", max_energy_fo_error)
        print("Max Energy Error for guiding center = ", max_energy_gc_error)
        print("Max mu Error for full orbit         = ", max_mu_fo_error)
        print("Max mu Error for guiding center     = ", max_mu_gc_error)
        assert max(max_energy_fo_error) < -8
        assert max(max_energy_gc_error) < -3
        assert max(max_mu_fo_error) < -3
        assert max(max_mu_gc_error) < -6

    def test_angularmomentum_conservation(self):
        # Test conservation of canonical angular momentum pphi
        # in an axisymmetric geometry where
        # pphi=q*poloidal_magnetic_flux+m*v_parallel*G/B
        R0test = 1.0
        B0test = 2.0
        qtest = 3.1
        bsh = ToroidalField(R0test, B0test)+PoloidalField(R0test, B0test, qtest)
        ma = CurveXYZFourier(300, 1)
        ma.set_dofs([0, 0, R0test+0.01, 0, R0test+0.01, 0., 0, 0., 0.])

        nparticles = 4
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        tmax = 1e-4
        Ekin = 100*ONE_EV
        np.random.seed(1)

        gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=-0.5, umax=-0.25,  # pitch angle so that we have both par and perp contribution
            phis=[], mode='gc_vac', tol=1e-11)
        if with_evtk:
            particles_to_vtk(gc_tys, '/tmp/particles_gc')

        # pick 100 random points on each trace

        N = 100
        max_pphi1_gc_error = np.array([])
        max_pphi2_gc_error = np.array([])
        max_pphi_gc_error = np.array([])
        for i in range(nparticles):
            gc_ty = gc_tys[i]
            idxs = np.random.randint(0, gc_ty.shape[0], size=(N, ))
            gc_xyzs = gc_ty[idxs, 1:4]
            bsh.set_points(gc_xyzs)
            AbsBs = bsh.AbsB()

            pphi1_gc = np.array([])
            pphi2_gc = np.array([])
            for j in range(N):
                v_gc = gc_ty[idxs[j], 4]
                r_squared = (np.sqrt(gc_xyzs[j][0]**2+gc_xyzs[j][1]**2)-R0test)**2+gc_xyzs[j][2]**2
                psi_poloidal = B0test*r_squared/qtest/2
                p_phi1 = q*psi_poloidal
                p_phi2 = m*v_gc*B0test*R0test/AbsBs[j]
                pphi1_gc = np.append(pphi1_gc, p_phi1)
                pphi2_gc = np.append(pphi2_gc, p_phi2)
            pphi_gc = pphi1_gc+pphi2_gc
            pphi1_error = np.log10(np.abs(pphi1_gc-pphi1_gc[0]+1e-30)/np.abs(pphi1_gc[0]))
            pphi2_error = np.log10(np.abs(pphi2_gc-pphi2_gc[0]+1e-30)/np.abs(pphi2_gc[0]))
            pphi_error = np.log10(np.abs(pphi_gc-pphi_gc[0]+1e-30)/np.abs(pphi_gc[0]))
            max_pphi1_gc_error = np.append(max_pphi1_gc_error, max(pphi1_error[3::]))
            max_pphi2_gc_error = np.append(max_pphi2_gc_error, max(pphi2_error[3::]))
            max_pphi_gc_error = np.append(max_pphi_gc_error, max(pphi_error[3::]))
        print("Max log10 variation of pphi1_gc = ", max_pphi1_gc_error)
        print("Max log10 variation of pphi2_gc = ", max_pphi2_gc_error)
        print("Max log10 variation of pphi1_gc + pphi2_gc = ", max_pphi_gc_error)
        assert max(max_pphi_gc_error) < -3

    def test_stopping_criteria(self):
        bsh = self.bsh
        ma = self.ma
        nparticles = 1
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        tmax = 1e-3
        Ekin = 9000*ONE_EV
        np.random.seed(1)

        gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=-0.80, umax=-0.70,
            phis=[], mode='gc_vac', tol=1e-11, stopping_criteria=[IterationStoppingCriterion(10)])
        assert len(gc_tys[0]) == 11

        # consider a particle with mostly perpendicular velocity so that it get's lost
        ntor = 1
        mpol = 1
        stellsym = True
        nfp = 3
        phis = np.linspace(0, 1, 50, endpoint=False)
        thetas = np.linspace(0, 1, 50, endpoint=False)
        s = SurfaceRZFourier(
            mpol=mpol, ntor=ntor, stellsym=stellsym, nfp=nfp, quadpoints_phi=phis, quadpoints_theta=thetas)
        s.fit_to_curve(ma, 0.03, flip_theta=False)
        sc = SurfaceClassifier(s, h=0.1, p=2)

        gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
            ma, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=-0.01, umax=+0.01,
            phis=[], mode='gc_vac', tol=1e-11, stopping_criteria=[LevelsetStoppingCriterion(sc)])
        if with_evtk:
            particles_to_vtk(gc_tys, '/tmp/particles_gc')
        assert gc_phi_hits[0][-1][1] == -1
        assert np.all(sc.evaluate(gc_tys[0][:, 1:4]) > 0)

    def test_tracing_on_surface_runs(self):
        bsh = self.bsh
        ma = self.ma
        nparticles = 1
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        tmax = 1e-3
        Ekin = 9000*ONE_EV
        np.random.seed(1)

        ntor = 1
        mpol = 1
        stellsym = True
        nfp = 3
        phis = np.linspace(0, 1, 50, endpoint=False)
        thetas = np.linspace(0, 1, 50, endpoint=False)
        s = SurfaceRZFourier(
            mpol=mpol, ntor=ntor, stellsym=stellsym, nfp=nfp, quadpoints_phi=phis, quadpoints_theta=thetas)
        s.fit_to_curve(ma, 0.03, flip_theta=False)

        gc_tys, gc_phi_hits = trace_particles_starting_on_surface(
            s, bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=-0.80, umax=-0.70,
            phis=[], mode='gc_vac', tol=1e-11, stopping_criteria=[IterationStoppingCriterion(10)])
        assert len(gc_tys[0]) == 11



#!/usr/bin/env python3

import numpy as np
from mpi4py import MPI

from simsopt import LeastSquaresProblem
from simsopt.mhd import Vmec
from simsopt.util.mpi import MpiPartition, log
from simsopt.solve.mpi import least_squares_mpi_solve

"""
This script implements the "1DOF_circularCrossSection_varyR0_targetVolume"
example from
https://github.com/landreman/stellopt_scenarios

This optimization problem has one independent variable, representing
the mean major radius. The problem also has one objective: the plasma
volume. There is not actually any need to run an equilibrium code like
VMEC since the objective function can be computed directly from the
boundary shape. But this problem is a fast way to test the
optimization infrastructure with VMEC.

Details of the optimum and a plot of the objective function landscape
can be found here:
https://github.com/landreman/stellopt_scenarios/tree/master/1DOF_circularCrossSection_varyR0_targetVolume
"""

# Print detailed logging info. This could be commented out if desired.
log()

# In the next line, we can adjust how many groups the pool of MPI
# processes is split into.
mpi = MpiPartition(ngroups=2)
mpi.write()

# Start with a default surface, which is axisymmetric with major
# radius 1 and minor radius 0.1.
equil = Vmec(mpi=mpi)
surf = equil.boundary

# Set the initial boundary shape. Here is one syntax:
surf.set('rc(0,0)', 1.0)
# Here is another syntax:
surf.set_rc(0, 1, 0.1)
surf.set_zs(0, 1, 0.1)

surf.set_rc(1, 0, 0.1)
surf.set_zs(1, 0, 0.1)

# VMEC parameters are all fixed by default, while surface parameters
# are all non-fixed by default.  You can choose which parameters are
# optimized by setting their 'fixed' attributes.
surf.all_fixed()
surf.set_fixed('rc(0,0)', False)

# Each Target is then equipped with a shift and weight, to become a
# term in a least-squares objective function.  A list of terms are
# combined to form a nonlinear-least-squares problem.
desired_volume = 0.15
prob = LeastSquaresProblem([(equil.volume, desired_volume, 1)])

# Solve the minimization problem. We can choose whether to use a
# derivative-free or derivative-based algorithm.
least_squares_mpi_solve(prob, mpi, grad=True)

# Make sure all procs call VMEC:
objective = prob.objective()
if mpi.proc0_world:
    print("At the optimum,")
    print(" rc(m=0,n=0) = ", surf.get_rc(0, 0))
    print(" volume, according to VMEC    = ", equil.volume())
    print(" volume, according to Surface = ", surf.volume())
    print(" objective function = ", objective)

assert np.abs(surf.get_rc(0, 0) - 0.7599088773175) < 1.0e-5
assert np.abs(equil.volume() - 0.15) < 1.0e-6
assert np.abs(surf.volume() - 0.15) < 1.0e-6
assert prob.objective() < 1.0e-15



#!/usr/bin/env python

from simsopt.util.mpi import MpiPartition, log
from simsopt.mhd import Vmec, Boozer, Quasisymmetry
from simsopt import LeastSquaresProblem
from simsopt.solve.mpi import least_squares_mpi_solve
import os

"""
This script solve the problem in
https://github.com/landreman/stellopt_scenarios/tree/master/7DOF_varyAxisAndElongation_targetIotaAndQuasisymmetry
See that website for a detailed description of the problem.
"""

# This next line turns on detailed logging. It can be commented out if
# you do not want such verbose output.
log()

mpi = MpiPartition()
vmec = Vmec(os.path.join(os.path.dirname(__file__), 'inputs', 'input.stellopt_scenarios_7dof'), mpi)

# We will optimize in the space of Garabedian coefficients:
surf = vmec.boundary.to_Garabedian()
vmec.boundary = surf

# Define parameter space:
surf.all_fixed()
surf.fixed_range(mmin=0, mmax=2, nmin=-1, nmax=1, fixed=False)
surf.set_fixed("Delta(1,0)")  # toroidally-averaged major radius
surf.set_fixed("Delta(0,0)")  # toroidally-averaged minor radius

# Define objective function:
boozer = Boozer(vmec, mpol=32, ntor=16)
qs = Quasisymmetry(boozer,
                   1.0,  # Radius to target
                   1, 0,  # (M, N) you want in |B|
                   normalization="symmetric",
                   weight="stellopt")

# Objective function is 100 * (iota - (-0.41))^2 + 1 * (qs - 0)^2
prob = LeastSquaresProblem([(vmec.iota_axis, -0.41, 100),
                            (qs, 0, 1)])

residuals = prob.f()
vals = prob.dofs.f()
if mpi.proc0_world:
    print("Initial values before shifting and scaling:  ", vals[:10])
    print("Initial residuals after shifting and scaling:", residuals[:10])
    print("size of residuals:", len(residuals))
    print("Initial objective function:", prob.objective())
    print("Parameter space:")
    for name in prob.dofs.names:
        print(name)
    print("Initial state vector:", prob.x)
    print("Initial iota on axis:", vmec.iota_axis())
#exit(0)

least_squares_mpi_solve(prob, mpi, grad=True)

residuals = prob.f()
vals = prob.dofs.f()
if mpi.proc0_world:
    print("Final values before shifting and scaling:  ", vals[:10])
    print("Final residuals after shifting and scaling:", residuals[:10])
    print("Final objective function:", prob.objective())
    print("Final state vector:", prob.x)
    print("Final iota on axis:", vmec.iota_axis())



import logging
import unittest
import numpy as np
try:
    from mpi4py import MPI
except:
    MPI = None
from simsopt._core.dofs import Dofs
from simsopt.objectives.functions import Beale
from simsopt.objectives.least_squares import LeastSquaresProblem

from simsopt.util.mpi import log
if MPI:
    from simsopt.util.mpi import MpiPartition  # , log
    from simsopt.solve.mpi import fd_jac_mpi, least_squares_mpi_solve


#log(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class TestFunction1():
    def __init__(self):
        self.x = np.array([1.2, 0.9, -0.4])
        self.fixed = np.full(3, False)

    def get_dofs(self):
        return self.x

    def set_dofs(self, x):
        self.x = x

    def J(self):
        return np.exp(self.x[0] ** 2 - np.exp(self.x[1]) + np.sin(self.x[2]))


class TestFunction2():
    def __init__(self):
        self.x = np.array([1.2, 0.9])
        self.fixed = np.full(2, False)

    def get_dofs(self):
        return self.x

    def set_dofs(self, x):
        self.x = x

    def f0(self):
        return np.exp(0 + self.x[0] ** 2 - np.exp(self.x[1]))

    def f1(self):
        return np.exp(1 + self.x[0] ** 2 - np.exp(self.x[1]))

    def f2(self):
        return np.exp(2 + self.x[0] ** 2 - np.exp(self.x[1]))

    def f3(self):
        return np.exp(3 + self.x[0] ** 2 - np.exp(self.x[1]))


class TestFunction3:
    """
    This is the Rosenbrock function again, but with some unnecessary
    MPI communication added in order to test optimization with MPI.
    """

    def __init__(self, comm):
        self.comm = comm
        self.x = [0., 0.]
        self.dummy = 42

    def get_dofs(self):
        return self.x

    def set_dofs(self, x):
        self.x = x

    def f0(self):
        # Do some random MPI stuff just for the sake of testing.
        self.comm.barrier()
        self.comm.bcast(self.x)
        return self.x[0] - 1

    def f1(self):
        # Do some random MPI stuff just for the sake of testing.
        self.comm.bcast(self.dummy)
        self.comm.barrier()
        return self.x[0] ** 2 - self.x[1]


@unittest.skipIf(MPI is None, "MPI is missing")
class SolveMpiTests(unittest.TestCase):
    def test_fd_jac_eval_points(self):
        """
        Check that fd_jac_mpi is evaluating the residual functions at the
        expected locations.
        """
        for ngroups in range(1, 2):
            mpi = MpiPartition(ngroups)
            b = Beale()  # Any Optimizable object with 2 d.o.f.'s will do.

            # First examine 1-sided differences
            prob = LeastSquaresProblem([(b, 0, 1)], diff_method="forward",
                                       abs_step=1e-6, rel_step=1e-2)

            b.set_dofs([0, 0.2])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[0.0, 1e-6, 0],
                                   [0.2, 0.2, 0.202]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

            b.set_dofs([0, 0])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[0.0, 1e-6, 0],
                                   [0.0, 0.0, 1e-6]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

            b.set_dofs([-3, -4])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[-3.0, -2.97, -3.0],
                                   [-4.0, -4.0, -3.96]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

            b.set_dofs([3e-7, 4e-7])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[3e-7, 1.3e-6, 3.0e-7],
                                   [4e-7, 4.0e-7, 1.4e-6]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

            # Now examine centered differences
            prob = LeastSquaresProblem([(b, 0, 1)], diff_method="centered",
                                       abs_step=1e-6, rel_step=1e-2)

            b.set_dofs([0, 0.2])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[1e-6, -1e-6, 0, 0],
                                   [0.2, 0.2, 0.202, 0.198]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

            b.set_dofs([0, 0])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[1e-6, -1e-6, 0, 0],
                                   [0, 0, 1e-6, -1e-6]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

            b.set_dofs([-3, -4])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[-2.97, -3.03, -3.00, -3.00],
                                   [-4.00, -4.00, -3.96, -4.04]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

            b.set_dofs([3e-7, 4e-7])
            jac, xs, evals = fd_jac_mpi(prob.dofs, mpi)
            xs_correct = np.array([[1.3e-6, -0.7e-6, 3.0e-7, 3.00e-7],
                                   [4.0e-7, 4.00e-7, 1.4e-6, -0.6e-6]])
            if mpi.proc0_groups:
                np.testing.assert_allclose(xs, xs_correct)

    def test_fd_jac(self):
        """
        Test the parallel finite-difference Jacobian calculation.
        """
        abs_step = 1.0e-7
        rel_step = 0
        for ngroups in range(1, 4):
            logger.debug('ngroups={}'.format(ngroups))
            mpi = MpiPartition(ngroups=ngroups)
            o = TestFunction1()
            d = Dofs([o], diff_method="forward", abs_step=abs_step, rel_step=rel_step)
            logger.debug('About to do worker loop 1')
            jac, xs, evals = fd_jac_mpi(d, mpi)
            jac_reference = np.array([[5.865176283537110e-01, -6.010834349701177e-01, 2.250910244305793e-01]])
            if mpi.proc0_world:
                np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            o.set_dofs(np.array([1.2, 0.9, -0.4]))
            jac = d.fd_jac()
            np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

            # Repeat with centered differences
            o.set_dofs(np.array([1.2, 0.9, -0.4]))
            logger.debug('About to do worker loop 2')
            d.diff_method = "centered"
            jac, xs, evals = fd_jac_mpi(d, mpi)
            jac_reference = np.array([[5.865175337071982e-01, -6.010834789627051e-01, 2.250910093037906e-01]])
            if mpi.proc0_world:
                np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            o.set_dofs(np.array([1.2, 0.9, -0.4]))
            jac = d.fd_jac()
            np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

            # Now try a case with different nparams and nfuncs.
            o = TestFunction2()
            d = Dofs([o.f0, o.f1, o.f2, o.f3], diff_method="forward",
                     abs_step=abs_step, rel_step=rel_step)
            logger.debug('About to do worker loop 3')
            jac, xs, evals = fd_jac_mpi(d, mpi)
            jac_reference = np.array([[8.657715439008840e-01, -8.872724499564555e-01],
                                      [2.353411054922816e+00, -2.411856577788640e+00],
                                      [6.397234502131255e+00, -6.556105911492693e+00],
                                      [1.738948636642590e+01, -1.782134355643450e+01]])
            if mpi.proc0_world:
                np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            o.set_dofs(np.array([1.2, 0.9]))
            jac = d.fd_jac()
            np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

            # Repeat with centered differences
            o.set_dofs(np.array([1.2, 0.9]))
            d.diff_method = "centered"
            logger.debug('About to do worker loop 4')
            jac, xs, evals = fd_jac_mpi(d, mpi)
            jac_reference = np.array([[8.657714037352271e-01, -8.872725151820582e-01],
                                      [2.353410674116319e+00, -2.411856754314101e+00],
                                      [6.397233469623842e+00, -6.556106388888594e+00],
                                      [1.738948351093228e+01, -1.782134486205678e+01]])
            if mpi.proc0_world:
                np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)
            # While we're at it, also test the serial FD Jacobian:
            o.set_dofs(np.array([1.2, 0.9]))
            jac = d.fd_jac()
            np.testing.assert_allclose(jac, jac_reference, rtol=1e-13, atol=1e-13)

    def test_fd_jac_abs_rel_steps(self):
        """
        Confirm that the parallel finite difference gradient gives nearly
        the same result regardless of whether absolute or relative
        steps are used.
        """
        rtol = 1e-6
        atol = 1e-6
        for ngroups in range(1, 4):
            for abs_step in [0, 1.0e-7]:
                # Only try rel_step=0 if abs_step is positive:
                rel_steps = [0, 1.0e-7]
                if abs_step == 0:
                    rel_steps = [1.0e-7]

                for rel_step in rel_steps:
                    for diff_method in ["forward", "centered"]:
                        logger.debug(f'ngroups={ngroups} abs_step={abs_step} ' \
                                     f'rel_step={rel_step} diff_method={diff_method}')
                        mpi = MpiPartition(ngroups=ngroups)
                        o = TestFunction1()
                        d = Dofs([o], diff_method=diff_method,
                                 abs_step=abs_step, rel_step=rel_step)
                        logger.debug('About to do worker loop 1')
                        jac, xs, evals = fd_jac_mpi(d, mpi)
                        jac_reference = np.array([[5.865175337071982e-01, -6.010834789627051e-01, 2.250910093037906e-01]])
                        if mpi.proc0_world:
                            np.testing.assert_allclose(jac, jac_reference, rtol=rtol, atol=atol)
                        # While we're at it, also test the serial FD Jacobian:
                        jac = d.fd_jac()
                        np.testing.assert_allclose(jac, jac_reference, rtol=rtol, atol=atol)

                        # Now try a case with different nparams and nfuncs.
                        o = TestFunction2()
                        d = Dofs([o.f0, o.f1, o.f2, o.f3], diff_method=diff_method,
                                 abs_step=abs_step, rel_step=rel_step)
                        logger.debug('About to do worker loop 2')
                        jac, xs, evals = fd_jac_mpi(d, mpi)
                        jac_reference = np.array([[8.657714037352271e-01, -8.872725151820582e-01],
                                                  [2.353410674116319e+00, -2.411856754314101e+00],
                                                  [6.397233469623842e+00, -6.556106388888594e+00],
                                                  [1.738948351093228e+01, -1.782134486205678e+01]])
                        if mpi.proc0_world:
                            np.testing.assert_allclose(jac, jac_reference, rtol=rtol, atol=atol)
                        # While we're at it, also test the serial FD Jacobian:
                        jac = d.fd_jac()
                        np.testing.assert_allclose(jac, jac_reference, rtol=rtol, atol=atol)

    def test_parallel_optimization_without_grad(self):
        """
        Test a full least-squares optimization.
        """
        for ngroups in range(1, 4):
            mpi = MpiPartition(ngroups=ngroups)
            o = TestFunction3(mpi.comm_groups)
            term1 = (o.f0, 0, 1)
            term2 = (o.f1, 0, 1)
            prob = LeastSquaresProblem([term1, term2])
            least_squares_mpi_solve(prob, mpi, grad=False)
            self.assertAlmostEqual(prob.x[0], 1)
            self.assertAlmostEqual(prob.x[1], 1)

    def test_parallel_optimization_with_grad(self):
        """
        Test a full least-squares optimization.
        """
        for ngroups in range(1, 4):
            for abs_step in [0, 1.0e-7]:
                # Only try rel_step=0 if abs_step is positive:
                rel_steps = [0, 1.0e-7]
                if abs_step == 0:
                    rel_steps = [1.0e-7]

                for rel_step in rel_steps:
                    for diff_method in ["forward", "centered"]:
                        logger.debug(f'ngroups={ngroups} abs_step={abs_step} ' \
                                     f'rel_step={rel_step} diff_method={diff_method}')
                        mpi = MpiPartition(ngroups=ngroups)
                        o = TestFunction3(mpi.comm_groups)
                        term1 = (o.f0, 0, 1)
                        term2 = (o.f1, 0, 1)
                        prob = LeastSquaresProblem([term1, term2], diff_method=diff_method,
                                                   abs_step=abs_step, rel_step=rel_step)
                        # Set initial condition different from 0,
                        # because otherwise abs_step=0 causes step
                        # size to be 0.
                        prob.x = [-0.1, 0.2]
                        least_squares_mpi_solve(prob, mpi)
                        self.assertAlmostEqual(prob.x[0], 1)
                        self.assertAlmostEqual(prob.x[1], 1)




import unittest
import os
import numpy as np
try:
    from mpi4py import MPI
except:
    MPI = None
try:
    import vmec
    vmec_found = True
except:
    vmec_found = False
from . import TEST_DIR

run_modes = {'all': 63,
             'input': 35,  # STELLOPT uses 35; V3FIT uses 7                                                                  
             'output': 8,
             'main': 45}   # STELLOPT uses 61; V3FIT uses 45                                                                 

success_codes = [0, 11]
reset_file = ''

#@unittest.skipIf(not vmec_found, "Valid Python interface to VMEC not found")


@unittest.skip("These tests are obsolete due to the refactoring of vmec.py and vmec_core.py")
class F90wrapVmecTests(unittest.TestCase):
    def setUp(self):
        """
        Set up the test fixture. This subroutine is automatically run
        by python's unittest module before every test.
        """
        self.fcomm = MPI.COMM_WORLD.py2f()
        rank = MPI.COMM_WORLD.Get_rank()
        self.verbose = (rank == 0)
        self.filename = os.path.join(TEST_DIR, 'input.li383_low_res')

        self.ictrl = np.zeros(5, dtype=np.int32)

        ier = 0
        numsteps = 0
        ns_index = -1
        iseq = rank
        self.ictrl[0] = 0
        self.ictrl[1] = ier
        self.ictrl[2] = numsteps
        self.ictrl[3] = ns_index
        self.ictrl[4] = iseq

        # Change the working directory to match the directory of this
        # file. Otherwise the vmec output files may be put wherever
        # the unit tests are run from.
        this_dir = os.path.dirname(__file__)
        if this_dir != '':
            os.chdir(os.path.dirname(__file__))

    def tearDown(self):
        """
        Tear down the test fixture. This subroutine is automatically
        run by python's unittest module before every test. Here we
        call runvmec with the cleanup flag to deallocate arrays, such
        that runvmec can be called again later.
        """
        self.ictrl[0] = 16  # cleanup
        vmec.runvmec(self.ictrl, self.filename, self.verbose, \
                     self.fcomm, reset_file)

    def test_read_input(self):
        """
        Try reading a VMEC input file.
        """
        self.ictrl[0] = run_modes['input']
        vmec.runvmec(self.ictrl, self.filename, self.verbose, \
                     self.fcomm, reset_file)

        self.assertTrue(self.ictrl[1] in success_codes)

        self.assertEqual(vmec.vmec_input.nfp, 3)
        self.assertEqual(vmec.vmec_input.mpol, 4)
        self.assertEqual(vmec.vmec_input.ntor, 3)
        print('rbc.shape:', vmec.vmec_input.rbc.shape)
        print('rbc:', vmec.vmec_input.rbc[101:103, 0:4])

        # n = 0, m = 0:
        self.assertAlmostEqual(vmec.vmec_input.rbc[101, 0], 1.3782)

        # n = 0, m = 1:
        self.assertAlmostEqual(vmec.vmec_input.zbs[101, 1], 4.6465E-01)

        # n = 1, m = 1:
        self.assertAlmostEqual(vmec.vmec_input.zbs[102, 1], 1.6516E-01)

    #@unittest.skip("This test won't work until a low-level issue with VMEC is fixed to allow multiple readins.")

    def test_run_read(self):
        """
        Try running VMEC, then reading in results from the wout file.
        """

        self.ictrl[0] = 1 + 2 + 4 + 8
        vmec.runvmec(self.ictrl, self.filename, self.verbose, \
                     self.fcomm, reset_file)

        self.assertTrue(self.ictrl[1] in success_codes)

        self.assertEqual(vmec.vmec_input.nfp, 3)
        self.assertEqual(vmec.vmec_input.mpol, 4)
        self.assertEqual(vmec.vmec_input.ntor, 3)
        print('rbc.shape:', vmec.vmec_input.rbc.shape)
        print('rbc:', vmec.vmec_input.rbc[101:103, 0:4])

        # n = 0, m = 0:
        self.assertAlmostEqual(vmec.vmec_input.rbc[101, 0], 1.3782)

        # n = 0, m = 1:
        self.assertAlmostEqual(vmec.vmec_input.zbs[101, 1], 4.6465E-01)

        # n = 1, m = 1:
        self.assertAlmostEqual(vmec.vmec_input.zbs[102, 1], 1.6516E-01)

        # Now try reading in the output
        wout_file = os.path.join(os.path.dirname(__file__), 'wout_li383_low_res.nc')
        ierr = 0
        vmec.read_wout_mod.read_wout_file(wout_file, ierr)
        self.assertEqual(ierr, 0)
        self.assertAlmostEqual(vmec.read_wout_mod.betatot, \
                               0.0426211525919469, places=4)

        print('iotaf.shape:', vmec.read_wout_mod.iotaf.shape)
        print('rmnc.shape:', vmec.read_wout_mod.rmnc.shape)

        self.assertAlmostEqual(vmec.read_wout_mod.iotaf[-1], \
                               0.6556508142482989, places=4)

        self.assertAlmostEqual(vmec.read_wout_mod.rmnc[0, 0], \
                               1.4760749266902973, places=4)


if __name__ == "__main__":
    unittest.main()



import unittest
import os
import logging
from pathlib import Path
import numpy as np

from simsopt.geo.surfacehenneberg import SurfaceHenneberg

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

try:
    from mpi4py import MPI
except:
    MPI = None

try:
    import vmec
    vmec_found = True
except ImportError:
    vmec_found = False

if (MPI is not None) and vmec_found:
    from simsopt.mhd.vmec import Vmec
else:
    Vmec = None

TEST_DIR = (Path(__file__).parent / ".." / "test_files").resolve()


class SurfaceHennebergTests(unittest.TestCase):
    def test_init_quadpoints(self):
        """
        Test that we can initialize the points either by an integer, list,
        or numpy array.
        """
        for phi_points in [32,
                           np.linspace(0.0, 1.0, 32, endpoint=False),
                           list(np.linspace(0.0, 1.0, 32, endpoint=False))]:

            for theta_points in [32,
                                 np.linspace(0.0, 1.0, 32, endpoint=False),
                                 list(np.linspace(0.0, 1.0, 32, endpoint=False))]:

                surf = SurfaceHenneberg(nfp=5, alpha_fac=-1, mmax=3, nmax=4,
                                        quadpoints_phi=phi_points,
                                        quadpoints_theta=theta_points)

    def test_names(self):
        """
        Check that the names of the dofs are set correctly.
        """
        surf = SurfaceHenneberg(nfp=1, alpha_fac=1, mmax=2, nmax=1)
        names_correct = ['R0nH(0)', 'R0nH(1)',
                         'Z0nH(1)',
                         'bn(0)', 'bn(1)',
                         'rhomn(0,1)',
                         'rhomn(1,-1)', 'rhomn(1,0)', 'rhomn(1,1)',
                         'rhomn(2,-1)', 'rhomn(2,0)', 'rhomn(2,1)']
        self.assertEqual(surf.names, names_correct)

    def test_set_get_dofs(self):
        """
        If the dofs are set to random numbers, and then you get the dofs,
        you should get what you set.
        """
        for mmax in range(1, 4):
            for nmax in range(4):
                surf = SurfaceHenneberg(nfp=1, alpha_fac=1, mmax=mmax, nmax=nmax)
                self.assertEqual(len(surf.names), len(surf.get_dofs()))

                # R0nH has nmax+1
                # Z0nH has nmax
                # bn has nmax + 1
                # rhomn for m=0 has nmax
                # rhomn for m>0 has 2 * nmax + 1
                nvals = nmax + 1 \
                    + nmax \
                    + nmax + 1 \
                    + nmax \
                    + mmax * (2 * nmax + 1)
                vals = np.random.rand(nvals) - 0.5
                surf.set_dofs(vals)
                np.testing.assert_allclose(surf.get_dofs(), vals)

    def test_set_get_rhomn(self):
        """
        Test get_rhomn() and set_rhomn(). If you set a particular mode and
        then get that mode, you should get what you set.
        """
        for mmax in range(1, 4):
            for nmax in range(4):
                surf = SurfaceHenneberg(nfp=2, alpha_fac=-1, mmax=mmax, nmax=nmax)
                for m in range(mmax + 1):
                    nmin = -nmax
                    if m == 0:
                        nmin = 1
                    for n in range(nmin, nmax + 1):
                        val = np.random.rand() - 0.5
                        surf.set_rhomn(m, n, val)
                        self.assertAlmostEqual(surf.get_rhomn(m, n), val)

    def test_fixed_range(self):
        """
        Test the fixed_range() function.
        """
        surf = SurfaceHenneberg(nfp=1, alpha_fac=1, mmax=2, nmax=1)
        # Order of elements:
        names_correct = ['R0nH(0)', 'R0nH(1)',
                         'Z0nH(1)',
                         'bn(0)', 'bn(1)',
                         'rhomn(0,1)',
                         'rhomn(1,-1)', 'rhomn(1,0)', 'rhomn(1,1)',
                         'rhomn(2,-1)', 'rhomn(2,0)', 'rhomn(2,1)']
        surf.fixed_range(20, 20, True)
        np.testing.assert_equal(surf.fixed, [True]*12)
        surf.fixed_range(20, 20, False)
        np.testing.assert_equal(surf.fixed, [False]*12)
        surf.fixed_range(0, 0, True)
        np.testing.assert_equal(surf.fixed, [True, False, False, False, False, False,
                                             False, False, False, False, False, False])
        surf.fixed_range(1, 0, True)
        np.testing.assert_equal(surf.fixed, [True, False, False, True, False, False,
                                             False, True, False, False, False, False])
        surf.fixed_range(2, 0, True)
        np.testing.assert_equal(surf.fixed, [True, False, False, True, False, False,
                                             False, True, False, False, True, False])
        surf.all_fixed(True)
        surf.fixed_range(0, 1, False)
        np.testing.assert_equal(surf.fixed, [False, False, False, True, True, False,
                                             True, True, True, True, True, True])
        surf.fixed_range(1, 1, False)
        np.testing.assert_equal(surf.fixed, [False, False, False, False, False, False,
                                             False, False, False, True, True, True])

    def test_indexing(self):
        """
        Check that set_rhomn() sets the expected element of the dofs vector.
        """
        mmax = 3
        nmax = 2
        surf = SurfaceHenneberg(nfp=2, alpha_fac=1, mmax=mmax, nmax=nmax)
        for n in range(nmax + 1):
            surf.R0nH[n] = 1000.0 + n
            surf.bn[n] = 2000.0 + n
        for n in range(1, nmax + 1):
            surf.Z0nH[n] = 3000.0 + n
        for m in range(mmax + 1):
            nmin = -nmax
            if m == 0:
                nmin = 1
            for n in range(nmin, nmax + 1):
                surf.set_rhomn(m, n, m * 100 + n)

        x = surf.get_dofs()
        np.testing.assert_allclose(x, [1.000e+03, 1.001e+03, 1.002e+03, 3.001e+03, 3.002e+03, 2.000e+03, 2.001e+03, \
                                       2.002e+03, 1.000e+00, 2.000e+00, 9.800e+01, 9.900e+01, 1.000e+02, 1.010e+02, \
                                       1.020e+02, 1.980e+02, 1.990e+02, 2.000e+02, 2.010e+02, 2.020e+02, 2.980e+02, \
                                       2.990e+02, 3.000e+02, 3.010e+02, 3.020e+02])
        self.assertListEqual(surf.names, ['R0nH(0)', 'R0nH(1)', 'R0nH(2)', 'Z0nH(1)', 'Z0nH(2)', 'bn(0)',
                                          'bn(1)', 'bn(2)', 'rhomn(0,1)', 'rhomn(0,2)', 'rhomn(1,-2)',
                                          'rhomn(1,-1)', 'rhomn(1,0)', 'rhomn(1,1)', 'rhomn(1,2)',
                                          'rhomn(2,-2)', 'rhomn(2,-1)', 'rhomn(2,0)', 'rhomn(2,1)',
                                          'rhomn(2,2)', 'rhomn(3,-2)', 'rhomn(3,-1)', 'rhomn(3,0)',
                                          'rhomn(3,1)', 'rhomn(3,2)'])

    def test_axisymm(self):
        """
        Verify to_RZFourier() for an axisymmetric surface with circular
        cross-section.
        """
        R0 = 1.5
        a = 0.3
        for nfp in range(1, 3):
            for alpha_fac in [-1, 0, 1]:
                for mmax in range(1, 3):
                    for nmax in range(3):
                        surfH = SurfaceHenneberg(nfp=nfp, alpha_fac=alpha_fac, mmax=mmax, nmax=nmax)
                        self.assertEqual(surfH.num_dofs(), len(surfH.get_dofs()))
                        surfH.R0nH[0] = R0
                        surfH.bn[0] = a
                        surfH.set_rhomn(1, 0, a)
                        surf = surfH.to_RZFourier()
                        for m in range(surf.mpol + 1):
                            for n in range(-surf.ntor, surf.ntor + 1):
                                if m == 0 and n == 0:
                                    self.assertAlmostEqual(surf.get_rc(m, n), R0)
                                    self.assertAlmostEqual(surf.get_zs(m, n), 0.0)
                                elif m == 1 and n == 0:
                                    self.assertAlmostEqual(surf.get_rc(m, n), a)
                                    self.assertAlmostEqual(surf.get_zs(m, n), a)
                                else:
                                    self.assertAlmostEqual(surf.get_rc(m, n), 0.0)
                                    self.assertAlmostEqual(surf.get_zs(m, n), 0.0)

    @unittest.skipIf(Vmec is None, "Valid Python interface to VMEC not found")
    def test_from_RZFourier(self):
        """
        Take a VMEC boundary in the original RZFourier representation,
        convert it to the Henneberg representation, and convert it
        back to the RZFourier representation. Coordinate-independent
        quantities like the volume and area should not change. If the
        result is converted to the Henneberg representation yet again,
        the degrees of freedom should be unchanged from the earlier
        Henneberg representation.
        """
        # Prepare some arbitrary points for testing gamma_lin:
        nlist = 15
        theta_list = np.linspace(-0.2, 0.8, nlist)
        phi_list = theta_list ** 2 - 0.5

        # Try a QI, QA, and a QH:
        files = ['input.W7-X_standard_configuration', 'input.cfqs_2b40', 'input.NuhrenbergZille_1988_QHS']
        alpha_facs = [1, 1, -1]

        for test_file, alpha_fac in zip(files, alpha_facs):
            vmec = Vmec(os.path.join(TEST_DIR, test_file))
            surf1 = vmec.boundary
            surf2 = SurfaceHenneberg.from_RZFourier(surf1, alpha_fac)
            self.assertEqual(surf2.num_dofs(), len(surf2.get_dofs()))
            surf3 = surf2.to_RZFourier()

            # Test gamma_lin:
            gamma2 = np.zeros((nlist, 3))
            gamma3 = np.zeros((nlist, 3))
            surf2.gamma_lin(gamma2, phi_list, theta_list)
            surf3.gamma_lin(gamma3, phi_list, theta_list)
            np.testing.assert_allclose(gamma2, gamma3, atol=1e-13, rtol=1e-13)

            # Test other surface functions:
            np.testing.assert_allclose(surf2.gamma(), surf3.gamma(), atol=1e-13, rtol=1e-13)
            np.testing.assert_allclose(surf2.gammadash2(), surf3.gammadash2(), atol=1e-12, rtol=1e-12)
            np.testing.assert_allclose(surf2.gammadash1(), surf3.gammadash1(), atol=1e-11, rtol=1e-11)
            np.testing.assert_allclose(surf1.volume(), surf2.volume(), atol=0, rtol=1e-3)
            np.testing.assert_allclose(surf1.volume(), surf3.volume(), atol=0, rtol=1e-3)
            np.testing.assert_allclose(surf1.area(), surf2.area(), atol=0, rtol=1e-3)
            np.testing.assert_allclose(surf1.area(), surf3.area(), atol=0, rtol=1e-3)
            surf4 = SurfaceHenneberg.from_RZFourier(surf3, alpha_fac, mmax=surf2.mmax, nmax=surf2.nmax)
            np.testing.assert_allclose(surf2.R0nH, surf4.R0nH, atol=1e-12, rtol=1e-4)
            np.testing.assert_allclose(surf2.Z0nH, surf4.Z0nH, atol=1e-12, rtol=1e-4)
            np.testing.assert_allclose(surf2.bn, surf4.bn, atol=1e-12, rtol=1e-4)
            np.testing.assert_allclose(surf2.rhomn, surf4.rhomn, atol=1e-8, rtol=1e-4)

    @unittest.skipIf(Vmec is None, "Valid Python interface to VMEC not found")
    def test_vmec(self):
        """
        If we run VMEC using a given boundary shape, then convert the
        boundary to the Henneberg representation and convert back and
        run VMEC again, the rotational transform should be almost
        identical.
        """
        vmec = Vmec(os.path.join(TEST_DIR, 'input.cfqs_2b40'))
        vmec.run()
        iota1 = vmec.wout.iotaf

        vmec.boundary = SurfaceHenneberg.from_RZFourier(vmec.boundary, 1)
        vmec.need_to_run_code = True
        vmec.run()
        iota2 = vmec.wout.iotaf

        logger.info(f'iota1: {iota1}')
        logger.info(f'iota2: {iota2}')
        logger.info(f'diff: {iota1 - iota2}')
        np.testing.assert_allclose(iota1, iota2, atol=1e-3, rtol=1e-3)
        # But if the 2 iota profiles are _exactly_ the same, vmec must
        # not have actually used the converted boundary.
        self.assertTrue(np.max(np.abs(iota1 - iota2)) > 1e-12)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import unittest

from simsopt.geo.curvexyzfourier import CurveXYZFourier, JaxCurveXYZFourier
from simsopt.geo.curverzfourier import CurveRZFourier
from simsopt.geo.curve import RotatedCurve
from simsopt.geo import parameters
from simsopt.geo.curveobjectives import CurveLength
from simsopt.objectives.graph_least_squares import LeastSquaresProblem
from simsopt.solve.graph_serial import least_squares_serial_solve

parameters['jit'] = False


class Testing(unittest.TestCase):

    def subtest_curve_length_optimisation(self, rotated):
        nquadrature = 100
        nfourier = 4
        nfp = 5
        curve = CurveRZFourier(nquadrature, nfourier, nfp, True)

        # Initialize the Fourier amplitudes to some random values
        x0 = np.random.rand(curve.dof_size) - 0.5
        x0[0] = 3.0
        curve.x = x0
        print('Initial curve dofs: ', curve.x)

        # Tell the curve object that the first Fourier mode is fixed, whereas
        # all the other dofs are not.
        curve.fix(0)

        if rotated:
            curve = RotatedCurve(curve, 0.5, flip=False)

        # Presently in simsgeo, the length objective is a separate object
        # rather than a function of Curve itself.
        obj = CurveLength(curve)

        # For now, we need to add this attribute to CurveLength. Eventually
        # this would hopefully be done in simsgeo, but for now I'll put it here.

        print('Initial curve length: ', obj.J())

        # Each target function is then equipped with a shift and weight, to
        # become a term in a least-squares objective function.
        # A list of terms are combined to form a nonlinear-least-squares
        # problem.
        prob = LeastSquaresProblem.from_tuples([(obj.J, 0.0, 1.0)])

        # At the initial condition, get the Jacobian two ways: analytic
        # derivatives and finite differencing. The difference should be small.
        # Ignoring Jacs for the time being
        # fd_jac = prob.dofs.fd_jac()
        # jac = prob.dofs.jac()
        # print('finite difference Jacobian:')
        # print(fd_jac)
        # print('Analytic Jacobian:')
        # print(jac)
        # print('Difference:')
        # print(fd_jac - jac)
        # assert np.allclose(fd_jac, jac, rtol=1e-4, atol=1e-4)

        # Solve the minimization problem:
        least_squares_serial_solve(prob, ftol=1e-10, xtol=1e-10, gtol=1e-10)

        print('At the optimum, x: ', prob.x)
        print(' Final curve dofs: ', curve.local_full_x)
        print(' Final curve length:    ', obj.J())
        print(' Expected final length: ', 2 * np.pi * x0[0])
        print(' objective function: ', prob.objective())
        assert abs(obj.J() - 2 * np.pi * x0[0]) < 1e-8

    def test_curve_first_derivative(self):
        for rotated in [True, False]:
            with self.subTest(rotated=rotated):
                self.subtest_curve_length_optimisation(rotated=rotated)


if __name__ == "__main__":
    unittest.main()
